# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module contains the PATH related utilities
inside the utils module
"""


import os
import shutil
import sys
from glob import glob
import datetime

from pyworkflow import SCIPION_SCRATCH, DOCSITEURLS
from pyworkflow.exceptions import PyworkflowException
from pyworkflow.config import Config

ROOT = "/"


def findFileRecursive(filename, path):
    for root, dirs, files in os.walk(path):
        if filename in files:
            return os.path.join(root, filename)
    return None   
        
        
def findFile(filename, *paths, **kwargs):
    """ Search if the file is present in some path in the *paths provided.
    Return None if not found.
    'recursive' can be passed in kwargs to iterate into subfolders.
    """
    recursive = kwargs.get('recursive', False)
    
    if filename:
        for p in paths:
            fn = os.path.join(p, filename)
            if os.path.exists(fn):
                return fn
            if recursive:
                f = findFileRecursive(filename, p)
                if f:
                    return f
    return None


def findRootFrom(referenceFile, searchFile):
    """ This method will find a path (root) from 'referenceFile'
    from which the 'searchFile' os.path.exists. 
    A practical example of 'referenceFile' is a metadata file
    and 'searchFile' is an image to be found from the metadata.
    Return None if the path is not found.
    """
    absPath = os.path.dirname(os.path.abspath(referenceFile))
    
    while absPath is not None and absPath != '/':
        if os.path.exists(os.path.join(absPath, searchFile)):
            return absPath
        absPath = os.path.dirname(absPath)
        
    return None   


def getParentFolder(path):
    """ Returns the absolute parent folder of a file or folder. Work for
    folders that ens with "/" which dirname can't"""
    return os.path.dirname(os.path.abspath(path))


def replaceExt(filename, newExt):
    """ Replace the current path extension(from last .)
    with a new one. The new one should not contains the ."""
    return os.path.splitext(filename)[0] + '.' + newExt


def replaceBaseExt(filename, newExt):
    """ Replace the current basename extension(from last .)
    with a new one. The new one should not contains the .
    """
    return replaceExt(os.path.basename(filename), newExt)


def removeBaseExt(filename):
    """Take the basename of the filename and remove extension"""
    return removeExt(os.path.basename(filename))


def removeExt(filename):
    """ Remove extension from basename """
    return os.path.splitext(filename)[0]


def joinExt(*extensions):
    """ Join several path parts with a ."""
    return '.'.join(extensions)


def getExt(filePath):
    """ Return the extension given a file. """
    return os.path.splitext(filePath)[1]


def cleanPath(*paths):
    """ Remove a list of paths, either folders or files"""
    for p in paths:
        if os.path.exists(p):
            if os.path.isdir(p):
                if os.path.islink(p):
                    os.remove(p)
                else:
                    shutil.rmtree(p)
            else:
                os.remove(p)


def cleanPattern(pattern):
    """ Remove all files that match the pattern. """
    files = glob(pattern)
    cleanPath(*files)


def copyPattern(pattern, destFolder):
    """ Copy all files matching the pattern to the given destination folder."""
    for file in glob(pattern):
        copyFile(file, destFolder)


def makePath(*paths):
    """ Create a list of paths if they don't os.path.exists.
    Recursively create all folder needed in a path.
    If a path passed is a file, only the directory will be created.
    """
    for p in paths:
        if not os.path.exists(p) and len(p):
            os.makedirs(p)


def makeTmpPath(protocol):
    """
    Create the scratch folder if SCIPION_SCRATCH variable is defined into the
    Scipion config, i.o.c create tmp folder
    """
    tmpPath = protocol._getTmpPath()
    if not os.path.exists(tmpPath) and len(tmpPath):
        scratchPath = Config.SCIPION_SCRATCH

        if scratchPath is None:  # Case when SCIPION_SCRATCH doesn't exist. TMP folder is created
            os.makedirs(tmpPath)
        else:
            try:
                project = protocol.getProject()
                folderId = "_".join([project.getShortName(),project.getProtWorkingDir(protocol)])
                tmpScratchFolder = os.path.join(scratchPath, folderId)
                if os.path.exists(tmpScratchFolder):
                    cleanPath(tmpScratchFolder)
                os.makedirs(tmpScratchFolder)  # Create scratch folder
                createAbsLink(tmpScratchFolder, tmpPath)  # Create a sym link

            except Exception as e:
                raise PyworkflowException("Couldn't create the temporary folder %s at:\n %s\nPlease, review %s variable." %
                                (folderId, scratchPath, SCIPION_SCRATCH), url=DOCSITEURLS.CONFIG_SECTION % "scratch-folder") from e


def makeFilePath(*files):
    """ Make the path to ensure that files can be written. """
    makePath(*[os.path.dirname(f) for f in files])


def missingPaths(*paths):
    """ Check if the list of paths os.path.exists.
    Will return the list of missing files,
    if the list is empty means that all path os.path.exists
    """
    return [p for p in paths if not os.path.exists(p)]


def getHomePath(user=''):
    """Return the home path of a give user."""
    return os.path.expanduser("~" + user)


def expandPattern(pattern, vars=True, user=True):
    """ Expand environment vars and user from a given pattern. """
    if vars:
        pattern = os.path.expandvars(pattern)
    if user:
        pattern = os.path.expanduser(pattern)
    return pattern


def getFiles(folderPath):
    """
    Gets all files of given folder and it subfolders.
    folderPath -- Folder path to get files.
    returns -- Set with all folder files.
    """
    filePaths = set()
    for path, dirs, files in os.walk(folderPath):
        for f in files:
            filePaths.add(os.path.join(path, f))
    return filePaths


def copyTree(source, dest):
    """
    Wrapper around the shutil.copytree, but allowing
    that the dest folder also os.path.exists.
    """
    if not os.path.exists(dest):
        shutil.copytree(source, dest, symlinks=True)
    else:
        for f in os.listdir(source):
            fnPath = os.path.join(source, f)
            if os.path.isfile(fnPath):
                shutil.copy(fnPath, dest)
            elif os.path.isdir(fnPath):
                copyTree(fnPath, os.path.join(dest, f))


def moveTree(src, dest):
    copyTree(src, dest)
    cleanPath(src)


def copyFile(source, dest):
    """ Shortcut to shutil.copy. """
    shutil.copy(source, dest)


def moveFile(source, dest):
    """ Move file from source to dest. """
    copyFile(source, dest)
    cleanPath(source)


def createLink(source, dest):
    """ Creates a relative link to a given file path. 
    Try to use common path for source and dest to avoid errors. 
    Different relative paths may exist since there are different valid paths
    for a file, it depends on the current working dir path"""
    if os.path.islink(dest):
        os.remove(dest)
        
    if os.path.exists(dest):
        raise Exception('Destination %s os.path.exists and is not a link'
                        % dest)
    sourcedir = getParentFolder(source)
    destdir = getParentFolder(dest)
    relsource = os.path.join(os.path.relpath(sourcedir, destdir),
                             os.path.basename(source))
    os.symlink(relsource, dest)


def createAbsLink(source, dest):
    """ Creates a link to a given file path"""
    if os.path.islink(dest):
        os.remove(dest)
        
    if os.path.exists(dest):
        raise Exception('Destination %s os.path.exists and is not a link' % dest)

    os.symlink(source, dest)


def getLastFile(pattern):
    """ Return the last file matching the pattern. """
    files = glob(pattern)
    if len(files):
        files.sort()
        return files[-1]
    return None


def commonPath(*paths):
    """ Return the common longest prefix path.
    It uses the python os.path.commonprefix and 
    then the direname over it since the former is
    implemented in char-by-char base.
    """
    return os.path.dirname(os.path.commonprefix(*paths))


# Console (and XMIPP) escaped colors, and the related tags that we create
# with Text.tag_config(). This dict is used in OutputText:addLine()
# See also http://www.termsys.demon.co.uk/vtansi.htm#colors
colorName = {'30': 'gray',
             '31': 'red',
             '32': 'green',
             '33': 'yellow',
             '34': 'blue',
             '35': 'magenta',
             '36': 'cyan',
             '37': 'white'}


def renderTextFile(fname, add, offset=0, lineNo=0, numberLines=True,
                   maxSize=400, headSize=40, tailSize=None, notifyLine=None, errors='strict'):
    """
    Call callback function add() on each fragment of text from file fname,
    delimited by lines and/or color codes.
      add: callback function add(txt, tag='normal')
      offset: byte offset - we start reading the file from there
      lineNo: lines will be numbered from this value on
      numberLines: whether to prepend the line numbers
    """
    textfile = open(fname, encoding='utf-8', errors=errors)
    size = (os.stat(fname).st_size - offset) / 1024  # in kB

    for line in iterBigFile(textfile, offset, size,
                            maxSize, headSize, tailSize):
        if line is not None:
            lineNo += 1
            if notifyLine is not None:
                notifyLine(line)
            renderLine(line, add, lineNo, numberLines)
        else:
            add("""\n
    ==> Too much data to read (%d kB) -- %d kB omitted
    ==> Click on """ % (size, size - headSize - (tailSize or headSize)))
            add(fname, 'link:%s' % fname)
            add(' to open it with the default viewer\n\n')
            if numberLines:
                add('    ==> Line numbers below are not '
                    'in sync with the input data\n\n')

    offset = textfile.tell()  # save last position in file
    textfile.close()

    return offset, lineNo


def renderLine(line, add, lineNo=1, numberLines=True):
    """
    Find all the fragments of formatted text in line and call
    add(fragment, tag) for each of them.
    """
    # Prepend line number
    if numberLines and lineNo:
        add('%05d:' % lineNo, 'cyan')
        add('   ')

    # iter 1\riter 2\riter 3  -->  iter 3
    if '\r' in line:
        line = line[line.rfind('\r')+1:]  # overwriting!

    # Find all console escape codes and use the appropriate tag instead.
    pos = 0  # current position in the line we are parsing
    attribute = None
    while True:
        # line looks like:
        #   'blah blah \x1b[{attr1};...;{attrn}mTEXT\x1b[0m blah blah'
        # where {attrn} is the color code (31 is red, for example). See
        # http://www.termsys.demon.co.uk/vtansi.htm#colors
        start = line.find('\x1b[', pos)
        if start < 0:  # no more escape codes, just add the remaining text
            add(line[pos:], attribute)
            break

        add(line[pos:start], attribute)
        end = line.find('m', start+2)
        if end < 0:  # an escape code interrupted by newline... weird
            break
        code = line[start+2:end]

        # See what attribute to use from now on, and update pos
        if code == '0':
            attribute = None
        else:
            attribute = colorName.get(code[-2:], None)
        pos = end + 1  # go to the character next to "m", the closing char


def iterBigFile(textfile, offset=0, size=None,
                maxSize=400, headSize=40, tailSize=None):
    """
    Yield lines from file textfile. If the size to read is bigger
    than maxSize then yield the first lines until headSize bytes, then
    yield None, then yield the last lines from tailSize bytes to the end.
    """
    if size is None:
        # Size in kB of the part of the file that we will read
        textfile.seek(0, 2)
        sizeKb = (textfile.tell() - offset) / 1024
    else:
        sizeKb = size

    headSizeB = headSize * 1024
    tailSizeB = (tailSize or headSize) * 1024

    textfile.seek(offset)
    # If the size is bigger than the max that we want to read (in kB).
    if 0 < maxSize < sizeKb:
        # maxSize <= 0 means we just want to read it all and not enter here.
        for line in textfile.read(headSizeB).split('\n'):
            yield line + '\n'
        yield None  # Special result to mark omitting lines
        textfile.seek(sizeKb*1024-tailSizeB)  # ready to show the last bytes

    # Add the remaining lines (from our last offset)
    for line in textfile:
        yield line


def createUniqueFileName(fn):
    """
    This function creates a file name that is similar to the original
    by adding a unique numeric suffix. check   NamedTemporaryFile
    from tempfile for alternatives
    """
    if not os.path.os.path.exists(fn):
        return fn

    path, name = os.path.split(fn)
    name, ext = os.path.splitext(name)

    make_fn = lambda i: os.path.join(path, '%s_tmp_%d_%s' % (name, i, ext))

    for i in range(2, sys.maxsize):
        uni_fn = make_fn(i)
        if not os.path.os.path.exists(uni_fn):
            return uni_fn

    return None


def getFileSize(fn):
    """ Shortcut to inspect the size of a file. """
    return os.stat(fn).st_size


def getFileLastModificationDate(fn):
    """ Returns the last modification date of a file or None
    if it doesn't exist. """
    if os.path.os.path.exists(fn):
        ts = os.path.getmtime(fn)
        return datetime.datetime.fromtimestamp(ts)
    else:
        print(fn + " does not exist!!. Can't check last modification date.")
        return None



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import sys

import pyworkflow as pw
from pyworkflow.project import Manager


def usage(error):
    print("""
    ERROR: %s
    
    Usage: scipion python scripts/clean_projects.py [SCIPION_USER_DATA] [--delete]
        Clean projects that has expired (now - creation time > life time)
        Optional to pass SCIPION_USER_DATA folder from which to read 'projects'.
        If --delete is not passed, only a message with the projects to be deleted
        will be shown. If used --delete, the projects will be deleted from filesystem.
    """ % error)
    sys.exit(1)


n = len(sys.argv)

if n > 3:
    usage("Incorrect number of input parameters")
    
delete = '--delete' in sys.argv

arg1 = sys.argv[1]
if n > 1 and arg1 != '--delete':
    customUserData = arg1
else:
    customUserData = pw.Config.SCIPION_USER_DATA

print("Loading projects from:\n %s" % customUserData)
 
# Create a new project
manager = Manager(workspace=customUserData)

for projInfo in manager.listProjects():
    projName = projInfo.getName()
    proj = manager.loadProject(projName)
    settings = proj.getSettings()
    
    leftTime = proj.getLeftTime()
    if (leftTime is not None and 
            leftTime.days < 0):
        if delete:
            print("Deleting: %s (%s) " % (projName, leftTime))
            manager.deleteProject(projName)
        else: 
            print("Should delete: %s (%s)" % (projName, leftTime))



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Laura del Cano         (ldelcano@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module handles process execution
"""

import sys
from subprocess import check_call
import psutil

from .utils import greenStr
from pyworkflow import Config


# The job should be launched from the working directory!
def runJob(log, programname, params,           
           numberOfMpi=1, numberOfThreads=1, 
           hostConfig=None, env=None, cwd=None, gpuList=None):

    command = buildRunCommand(programname, params, numberOfMpi, hostConfig,
                              env, gpuList=gpuList)
    
    if log is None:
        print("** Running command: %s" % greenStr(command))
    else:
        log.info(greenStr(command))

    return runCommand(command, env, cwd)
        

def runCommand(command, env=None, cwd=None):
    """ Execute command with given environment env and directory cwd """

    # First let us create core dumps if in debug mode
    if Config.debugOn(env):
        import resource
        resource.setrlimit(resource.RLIMIT_CORE,
                           (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        # This is like "ulimit -u 99999999", so we can create core dumps

    # TODO: maybe have to set PBS_NODEFILE in case it is used by "command"
    # (useful for example with gnu parallel)
    check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr,
               env=env, cwd=cwd)
    # It would be nice to avoid shell=True and calling buildRunCommand()...

    
def buildRunCommand(programname, params, numberOfMpi, hostConfig=None,
                    env=None, gpuList=None):
    """ Return a string with the command line to run """

    # Convert our list of params to a string, with each element escaped
    # with "" in case there are spaces.
    if not isinstance(params, str):
        params = ' '.join('"%s"' % p for p in params)

    if gpuList:
        params = params % {'GPU': ' '.join(str(g) for g in gpuList)}

    prepend = '' if env is None else env.getPrepend()

    if numberOfMpi <= 1:
        return '%s %s %s' % (prepend, programname, params)
    else:
        assert hostConfig is not None, 'hostConfig needed to launch MPI processes.'

        if programname.startswith('xmipp') and not programname.startswith('xmipp_mpi'):
            programname = programname.replace('xmipp', 'xmipp_mpi')
            
        mpiFlags = '' if env is None else env.get('SCIPION_MPI_FLAGS', '') 

        mpiCmd = hostConfig.mpiCommand.get() % {
            'JOB_NODES': numberOfMpi,
            'COMMAND': "%s `which %s` %s" % (mpiFlags, programname, params),
        }
        return '%s %s' % (prepend, mpiCmd)


def killWithChilds(pid):
    """ Kill the process with given pid and all children processes.
    Params:
     pid: the process id to terminate
    """
    proc = psutil.Process(pid)
    for c in proc.children(recursive=True):
        if c.pid is not None:
            print("Terminating child pid: %d" % c.pid)
            c.kill()
    print("Terminating process pid: %s" % pid)
    if pid is None:
        print("WARNING! Got None PID!!!")
    else:
        proc.kill()


def isProcessAlive(pid):
    try:
        psutil.Process(pid)
        return True
    except Exception:
        return False



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Tree widget implementation.
"""
        
import tkinter as tk

from pyworkflow.project import Label
from pyworkflow.gui import Icon, configureWeigths
from pyworkflow.gui.tree import TreeProvider
import pyworkflow.gui.dialog as dialog


class LabelsTreeProvider(TreeProvider):
    """ Populate Tree from Labels. """
    def __init__(self, objList=None):
        TreeProvider.__init__(self)
        self.objList = objList
        self._parentDict = {}

    def getColumns(self):
        return [('name', 300), ('color', 150)]

    def getObjectInfo(self, label):

        return {'key': label.getId(), 'parent': None,
                'text': label.getName(), 'values': (label.getColor()),
                'tags': label.getColor()}

    def getObjectPreview(self, obj):
        return None, None

    def getObjectActions(self, obj):
        return []

    def _getObjectList(self):
        """Retrieve the object list"""
        return self.objList

    def getObjects(self):
        objList = self._getObjectList()
        return objList

    def configureTags(self, tree):
        for label in self.getObjects():
            tree.tag_configure(label.getColor(), background=label.getColor())


class LabelsDialog(dialog.ToolbarListDialog):
    """
    This class extend from ListDialog to allow an
    extra toolbar to handle operations over the elements
    in the list (e.g. Edit, New, Delete).
    """
    def __init__(self, parent, labels, **kwargs):
        """ From kwargs:
                message: message tooltip to show when browsing.
                selected: the item that should be selected.
                validateSelectionCallback:
                    a callback function to validate selected items.
                allowSelect: if set to False, the 'Select' button will not
                    be shown.
        """
        self.labels = labels
        toolbarButtons = [
            dialog.ToolbarButton('Add', self._addLabel, Icon.ACTION_NEW),
            dialog.ToolbarButton('Edit', self._editLabel, Icon.ACTION_EDIT),
            dialog.ToolbarButton('Delete', self._deleteLabel, Icon.ACTION_DELETE)
        ]

        dialog.ToolbarListDialog.__init__(self, parent,
                                          "Manage labels",
                                          LabelsTreeProvider(labels),
                                          "Select the label to edit or delete",
                                          toolbarButtons,
                                          allowsEmptySelection=True,
                                          itemDoubleClick=self._editLabel,
                                          **kwargs)

    def _newColor(self):
        """ Pick a color by default for a given label from a predefined list.
         Check that the color have not been used by other label.
        """
        colors = ["#e57373", "#4fc3f7", "#81c784", "#ff8a65", "#9575cd",
                  "#a1887f", "#ffd54f", "#dce775", "#4db6ac"]

        for c in colors:
            if all(l.getColor().lower() != c for l in self.labels):
                return c

        return 'red'

    def _addLabel(self, e=None):
        label = Label(color=self._newColor())
        dlg = EditLabelDialog(self, "Add label", label)
        if dlg.resultYes():
            self.labels.addLabel(label)
            self.tree.update()

    def _editLabel(self, e=None):
        selection = self.tree.getSelectedObjects()
        if selection:
            label = selection[0]
            dlg = EditLabelDialog(self, "Edit label", label)
            if dlg.resultYes():
                self.tree.update()

    def _deleteLabel(self, e=None):
        selection = self.tree.getSelectedObjects()
        if selection:
            labelsStr = '\n'.join('- %s' % l.getName() for l in selection)
            if dialog.askYesNo("Delete a label",
                               "Are you sure to delete the "
                               "following label(s)?\n %s" % labelsStr, self):
                for label in selection:
                    self.labels.deleteLabel(label)
                self.tree.update()


class EditLabelDialog(dialog.Dialog):
    """ Dialog to edit a label (name, color) """
    def __init__(self, parent, title, label, **kwargs):
        self.label = label
        dialog.Dialog.__init__(self, parent, title)

    def body(self, bodyFrame):
        bodyFrame.config(bg='white')
        configureWeigths(bodyFrame, 1, 1)

        # Label
        label_text = tk.Label(bodyFrame, text="Name", bg='white', bd=0)
        label_text.grid(row=0, column=0, sticky='nw', padx=(15, 10), pady=15)
        # Label box
        var = tk.StringVar()
        var.set(self.label.getName())
        self.textVar = var
        self.textLabel = tk.Entry(bodyFrame, width=20, textvariable=var)
        self.textLabel.grid(row=0, column=1, sticky='news', padx=5, pady=5)

        # Comment
        colorLabel = tk.Label(bodyFrame, text='Color \n(Click to change)',
                              bg='white', bd=0)
        colorLabel.grid(row=1, column=0, sticky='nw', padx=(15, 10), pady=15)
        self.colorVar = tk.StringVar()
        self.colorVar.set(self.label.getColor())
        self.colorBox = tk.Frame(bodyFrame, bg=self.colorVar.get())
        self.colorBox.grid(row=1, column=1, sticky='news', padx=5, pady=5)
        colorLabel.bind('<Button-1>', self._changeColor)
        self.colorBox.bind('<Button-1>', self._changeColor)

    def apply(self):
        self.label.setName(self.textVar.get())
        self.label.setColor(self.colorVar.get())

    def _changeColor(self, e=None):
        hexColor = dialog.askColor(parent=self, defaultColor=self.colorVar.get())
        if hexColor is not None:
            self.colorBox.config(bg=hexColor)
            self.colorVar.set(hexColor)

    def validate(self):

        validationMsg = None

        if len(self.textVar.get().strip()) == 0:
            validationMsg = "Label name can't be empty.\n"

        if validationMsg is not None:
            dialog.showError("Validation error", validationMsg, self)
            return False

        return True






# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Scipion team
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules contains constants related to Pyworkflow
"""
# This variable is useful to determinate the plugins compatibility with the
# current Scipion core release.
# This version does not need to change with future scipion releases
# if plugins are still compatible, so future hot fixes releases or even micros
# or minor release should not change this CORE_VERSION. Only, when a new release
# will break existing plugins, this number needs to be incremented.
CORE_VERSION = '3.0.0'

# Versions
VERSION_1 = '1.0.0'
VERSION_1_1 = '1.1.0'
VERSION_1_2 = '1.2.0'
VERSION_2_0 = '2.0.0'
VERSION_3_0 = '3.0.21'

# For a new release, define a new constant and assign it to LAST_VERSION
# The existing one has to be added to OLD_VERSIONS list.
LAST_VERSION = VERSION_3_0
OLD_VERSIONS = (VERSION_1, VERSION_1_1, VERSION_1_2, VERSION_2_0)

# STATUS
PROD = 0
BETA = 1
NEW = 2

# Dir names
APPS = 'apps'
DATA_TAG = 'data'

# Scripts
PW_SYNC_DATA = 'pw_sync_data.py'
PW_SCHEDULE_RUN = 'pw_schedule_run.py'
PW_PROTOCOL_MPIRUN = 'pw_protocol_mpirun.py'
PW_RUN_TESTS = 'pw_run_tests.py'
PW_VIEWER = 'pw_viewer.py'

# PW Config
SCIPION_PYTHON = 'SCIPION_PYTHON'
SCIPION_PYTHON_DEFAULT = 'python3'
SCIPION_HOME_VAR = 'SCIPION_HOME'
SCIPION_TESTS = 'SCIPION_TESTS'
SCIPION_SCRATCH = 'SCIPION_SCRATCH'

# NOTES CONSTANTS
SCIPION_NOTES_FILE = 'SCIPION_NOTES_FILE'
SCIPION_NOTES_FILE_DEFAULT = 'notes.txt'
SCIPION_NOTES_PROGRAM = 'SCIPION_NOTES_PROGRAM'
SCIPION_NOTES_ARGS = 'SCIPION_NOTES_ARGS'
SCIPION_NOTES_HEADING_MSG = \
    '############################################  SCIPION NOTES  ##############################################' \
    '\n\nThis document can be used to store your notes within your project from Scipion framework.\n\n' \
    'Scipion notes behaviour can be managed in the Scipion config file by creating or editing, if they\n' \
    'already exist, the following variables:\n\n' \
    '\t-%s is used to store the file name (default is %s)\n' \
    '\t-%s is used to select the program which will be used to open the notes file. If \n' \
    '\t empty, it will use the default program used by your OS to open that type of file.\n' \
    '\t-%s is used to add input arguments that will be used in the calling of the program\n' \
    '\t specified in %s.\n\n' \
    'These lines can be removed if desired.\n\n' \
    '###########################################################################################################' \
    '\n\nPROJECT NOTES:' % (SCIPION_NOTES_FILE, SCIPION_NOTES_FILE_DEFAULT,
                            SCIPION_NOTES_PROGRAM, SCIPION_NOTES_ARGS,
                            SCIPION_NOTES_PROGRAM)

SCIPION_DOMAIN = 'SCIPION_DOMAIN'

# Debug constants
SCIPION_DEBUG = 'SCIPION_DEBUG'
SCIPION_JSON_TEMPLATES = '.json.template'
SCIPION_DEBUG_NOCLEAN = 'SCIPION_DEBUG_NOCLEAN'
SCIPION_DEBUG_SQLITE = 'SCIPION_DEBUG_SQLITE'
SCIPION_LOG_LEVEL = 'SCIPION_LOG_LEVEL'

# Color and appearance constants
TK_GRAY_DEFAULT = '#d9d9d9'

# Other
SCIPION_TESTS_CMD = 'SCIPION_TESTS_CMD'
CONDA_ACTIVATION_CMD_VAR = 'CONDA_ACTIVATION_CMD'
VIEWERS = 'VIEWERS'

# Results when updating a protocol
NOT_UPDATED_READ_ONLY = 0
NOT_UPDATED_UNNECESSARY = 1
NOT_UPDATED_ERROR = 2
PROTOCOL_UPDATED = 3

# Db names
PROJECT_DBNAME = 'project.sqlite'
PROJECT_SETTINGS = 'settings.sqlite'



class DOCSITEURLS:
    """Documentation site URL useful when exceptions happens and you want to point to some pages"""
    HOME = 'https://scipion-em.github.io/docs/'
    DOCS = HOME + 'docs/'
    CONFIG = DOCS + 'scipion-modes/scipion-configuration.html'
    CONFIG_SECTION = CONFIG + '#%s'
    CONTACTUS = 'http://scipion.i2pc.es/contact'
    USER = DOCS + 'user/'
    PLUGIN_MANAGER = USER + 'plugin-manager.html'
    HOST_CONFIG = DOCS + "scipion-modes/host-configuration.html"

# tkinter bind constants
class TK:
    LEFT_CLICK = '<Button-1>'
    RETURN = '<Return>'
    ENTER = '<KP_Enter>'
    LEFT_DOUBLE_CLICK = '<Double-1>'
    TREEVIEW_OPEN = '<<TreeviewOpen>>'
    TREEVIEW_CLOSE = '<<TreeviewClose>>'



# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import os
import time

import pyworkflow as pw
from pyworkflow.project import Manager
from pyworkflow.project import Project
import pyworkflow.utils as pwutils
from pyworkflow.protocol import INITIAL_SLEEP_TIME


def usage(error):
    print("""
    ERROR: %s
    
    Usage: python -m pyworkflow.project.scripts.schedule projectName 
    
              options: --ignore ProtClassName1 ProtClassName2 ProtClassLabel1 ...
              
        This script will schedule all the protocols in a project. If '--ignore' 
          options is passed, it doesn't schedule those protocols that belongs to 
          ProtClassName1 or ProtClassName2 class, also those protocols with a 
          objLabel equals to ProtClassLabel1
    """ % error)
    sys.exit(1)


n = len(sys.argv)

if n < 2:
    usage("This script accepts 1 mandatory parameter: the project name.")
elif n > 2 and sys.argv[2] != '--ignore':
    usage("The protocol class names to be ignored must be after a '--ignore' flag.")

projName = sys.argv[1]

# This fails, since it is triggering matplotlib.pyplot and then import error happens:
# ... pyworkflow/gui/no-tkinter/_tkinter.py: invalid ELF header. If we want this back we might need to
# invest some time "faking" tkinter again for python3.
# path = pw.join('gui', 'no-tkinter')
# sys.path.insert(1, path)

# Create a new project
manager = Manager()

if not manager.hasProject(projName):
    usage("There is no project with this name: %s"
          % pwutils.red(projName))

# the project may be a soft link which may be unavailable to the cluster
# so get the real path

try:
    projectPath = os.readlink(manager.getProjectPath(projName))
except:
    projectPath = manager.getProjectPath(projName)

project = Project(pw.Config.getDomain(), projectPath)
project.load()

runGraph = project.getRunsGraph()
roots = runGraph.getRootNodes()

# Now assuming that there is no dependencies between runs
# and the graph is lineal

for root in roots:
    for child in root.getChilds():
        workflow, _ = project._getWorkflowFromProtocol(child.run)
        for prot, level in workflow.values():
            if prot.forceSchedule.get():
                print(pwutils.blueStr("\nStopping scheduling at '%s'. Force scheduling found.\n"
                                      % protLabelName))
                sys.exit()

            protClassName = prot.getClassName()
            protLabelName = prot.getObjLabel()
            if (protClassName not in sys.argv[3:] and
                    protLabelName not in sys.argv[3:]):
                project.scheduleProtocol(prot,
                                         initialSleepTime=level*INITIAL_SLEEP_TIME)
            else:
                print(pwutils.blueStr("\nNot scheduling '%s' protocol named '%s'.\n"
                                      % (protClassName, protLabelName)))



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module extends the functionalities of a normal Tkinter Canvas.
The new Canvas class allows to easily display Texboxes and Edges
that can be interactively dragged and clicked.
"""
import math
import tkinter as tk
import operator

from pyworkflow.gui import gui, getDefaultFont, cfgFontSize
from pyworkflow.gui.widgets import Scrollable

DEFAULT_ZOOM = 100

DEFAULT_FONT_SIZE = cfgFontSize

DEFAULT_CONNECTOR_FILL = "blue"
DEFAULT_CONNECTOR_OUTLINE = "black"


class Canvas(tk.Canvas, Scrollable):
    """Canvas to draw some objects.
    It will really contains a Frame, a Canvas and scrollbars"""
    _images = {}

    def __init__(self, parent, tooltipCallback=None, tooltipDelay=1500, **kwargs):
        defaults = {'bg': 'white'}
        defaults.update(kwargs)
        Scrollable.__init__(self, parent, tk.Canvas, **defaults)

        self.lastItem = None  # Track last item selected
        self.lastPos = (0, 0)  # Track last clicked position
        self.eventPos = (0, 0)
        self.firstPos = None  # Track first clicked position (for a drag action)
        self.items = {}  # Keep a dictionary with high-level items
        self.cleanSelected = True

        self.onClickCallback = None
        self.onDoubleClickCallback = None
        self.onRightClickCallback = None
        self.onControlClickCallback = None
        self.onAreaSelected = None

        # Add bindings
        self.bind("<Button-1>", self.onClick)
        self.bind("<ButtonRelease-1>", self.onButton1Release)
        self.bind("<Button-3>", self.onRightClick)
        self.bind("<Button-2>", self.onRightClick)
        self.bind("<Double-Button-1>", self.onDoubleClick)
        self.bind("<B1-Motion>", self.onDrag)
        # Hide the right-click menu
        self.bind('<FocusOut>', self._unpostMenu)
        self.bind("<Key>", self._unpostMenu)
        self.bind("<Control-1>", self.onControlClick)
        # self.bind("<MouseWheel>", self.onScroll)
        # Scroll bindings in Linux
        self.bind("<Shift-Button-4>", self.zoomerP)
        self.bind("<Shift-Button-5>", self.zoomerM)

        self._tooltipId = None
        self._tooltipOn = False  # True if the tooltip is displayed
        self._tooltipCallback = tooltipCallback
        self._tooltipDelay = tooltipDelay

        self._runsFont = getDefaultFont().copy()
        self._zoomFactor = DEFAULT_ZOOM
        self.nodeList = None

        if tooltipCallback:
            self.bind('<Motion>', self.onMotion)
            # self.bind('<Leave>', self.onLeave)
            self._createTooltip()  # This should set

        self._menu = tk.Menu(self, tearoff=0)

    def _createTooltip(self):
        """ Create a Tooltip window to display tooltips in 
        the canvas.
        """
        tw = tk.Toplevel(self)
        tw.withdraw()  # hidden by default
        tw.wm_overrideredirect(1)  # Remove window decorations
        tw.bind("<Leave>", self.hideTooltip)

        self._tooltip = tw

    def _showTooltip(self, x, y, item):
        # check that the mouse is still in the position
        nx = self.winfo_pointerx()
        ny = self.winfo_pointery()
        if x == nx and y == ny:
            self._tooltipOn = True
            tw = self._tooltip  # short notation
            self._tooltipCallback(tw, item)
            tw.update_idletasks()
            tw.wm_geometry("+%d+%d" % (x, y))
            tw.deiconify()

    def hideTooltip(self, e=None):
        if self._tooltipOn:
            self._tooltipOn = False
            tw = self._tooltip  # short notation
            tw.withdraw()

    def getRunsFont(self):
        return self._runsFont

    def getImage(self, img):
        return gui.getImage(img, self._images)

    def _unpostMenu(self, e=None):
        self._menu.unpost()

    def getCoordinates(self, event):
        """Converts the events coordinates to canvas coordinates"""
        # Convert screen coordinates to canvas coordinates
        xc = self.canvasx(event.x)
        yc = self.canvasy(event.y)
        return xc, yc

    def selectItem(self, item):
        if self.lastItem:
            self.lastItem.setSelected(False)
        self.lastItem = item
        item.setSelected(True)

    def multipleItemsSelected(self):
        """ Returns True if more than one box selected,
        False otherwise.
        TODO: add numItemsSelected as attribute to Canvas
        class and update when selection changes
        """
        selectedItemCounts = 0
        for k, v in self.items.items():
            if v.getSelected():
                selectedItemCounts += 1
            if selectedItemCounts > 1:
                return True
        return False

    def _findItem(self, xc, yc):
        """ Find if there is any item in the canvas
        in the mouse event coordinates.
        Return None if not Found
        """
        items = self.find_overlapping(xc - 1, yc - 1, xc + 1, yc + 1)
        if self.lastItem is not None and self.lastItem.id in items:
            return self.lastItem
        for i in items:
            if i in self.items:
                return self.items[i]
        return None

    def _handleMouseEvent(self, event, callback):
        # Store last event coordinates
        self.eventPos = (event.x, event.y)
        # Retrieve the coordinates relative to the Canvas
        xc, yc = self.getCoordinates(event)
        self.lastItem = self._findItem(xc, yc)
        self.callbackResults = None
        self.lastPos = (0, 0)

        if self.lastItem is not None:
            if callback:
                self.callbackResults = callback(self.lastItem)
        self.lastPos = (xc, yc)

    def onClick(self, event):
        self.cleanSelected = True
        self._unpostMenu()
        self._handleMouseEvent(event, self.onClickCallback)
        if self.lastItem is None:
            self.move_start(event)

    def onControlClick(self, event):
        self.cleanSelected = False
        self._unpostMenu()
        self._handleMouseEvent(event, self.onControlClickCallback)

    def onRightClick(self, e=None):
        # RightClick callback will not work not, as it need
        # the event information to know the coordinates
        self._handleMouseEvent(e, self.onRightClickCallback)
        unpost = True
        # If the callback return a list of actions
        # we will show up a menu with them
        actions = self.callbackResults

        if actions:
            self._menu.delete(0, tk.END)
            for a in actions:
                if a is None:
                    self._menu.add_separator()
                else:
                    img = ''
                    if len(a) > 2:  # image for the action
                        img = self.getImage(a[2])
                    self._menu.add_command(label=a[0], command=a[1],
                                           image=img, compound=tk.LEFT,
                                           font=gui.getDefaultFont())
            self._menu.post(e.x_root, e.y_root)
            unpost = False
        if unpost:
            self._menu.unpost()

    def onDoubleClick(self, event):
        self._handleMouseEvent(event, self.onDoubleClickCallback)

    # move
    def move_start(self, event):
        # If nothing was click on ButtonPress
        if self.lastItem is None:
            self.config(cursor='fleur')
            self.scan_mark(event.x, event.y)

    def onDrag(self, event):
        try:

            if self.lastItem:
                xc, yc = self.getCoordinates(event)
                self.lastItem.move(xc - self.lastPos[0], yc - self.lastPos[1])
                self.lastPos = (xc, yc)
            else:
                if self.firstPos is None:
                    self.firstPos = (event.x, event.y)
                (x, y) = self.getCoordinates(event)
                self.scan_dragto(event.x, event.y, gain=1)

        except Exception as ex:
            # JMRT: We are having a weird exception here.
            # Presumably because there is concurrency between the onDrag
            # event and the refresh one. For now, just ignore it.
            pass

    def onButton1Release(self, event):
        if self.firstPos is not None:
            # Failing in "data" view.
            # TODO: This is not fully implemented
            # self.onAreaSelected(self.firstPos[0], self.firstPos[1], event.x, event.y)

            self.firstPos = None

        self.config(cursor='arrow')

    def onMotion(self, event):
        self.onLeave(event)  # Hide tooltip and cancel schedule

        xc, yc = self.getCoordinates(event)
        item = self._findItem(xc, yc)
        if item is not None:
            self._tooltipId = self.after(self._tooltipDelay,
                                         lambda: self._showTooltip(event.x_root,
                                                                   event.y_root,
                                                                   item))

    def onLeave(self, event):
        if self._tooltipId:
            self.after_cancel(self._tooltipId)
            self.hideTooltip()

    def createTextbox(self, text, x, y, bgColor="#99DAE8", textColor='black'):
        tb = TextBox(self, text, x, y, bgColor, textColor)
        self.items[tb.id] = tb
        return tb

    def createTextCircle(self, text, x, y, bgColor="#99DAE8", textColor='black'):
        tb = TextCircle(self, text, x, y, bgColor, textColor)
        self.items[tb.id] = tb
        return tb

    def createRoundedTextbox(self, text, x, y, bgColor="#99DAE8", textColor='black'):
        tb = RoundedTextBox(self, text, x, y, bgColor, textColor)
        self.items[tb.id] = tb
        return tb

    def addItem(self, item):
        self.items[item.id] = item

    def createEdge(self, srcItem, dstItem):
        edge = Edge(self, srcItem, dstItem)
        # self.items[edge.id] = edge
        return edge

    def createCable(self, src, srcSocket, dst, dstSocket):
        return Cable(self, src, srcSocket, dst, dstSocket)

    def clear(self):
        """ Clear all items from the canvas """
        self.delete(tk.ALL)
        self.items.clear()

    def updateScrollRegion(self):
        self.update_idletasks()
        self.config(scrollregion=self.bbox("all"))

    # linux zoom
    def __zoom(self, event, scale):

        newZoomFactor = round(self._zoomFactor * scale)

        if self._zoomFactor == newZoomFactor:
            return

        self._zoomFactor = newZoomFactor

        # x, y = self.getCoordinates(event)
        self.scale("all", 0, 0, scale, scale)
        self.__scaleFonts()
        self.configure(scrollregion=self.bbox("all"))

    def __scaleFonts(self):

        currentFontSize = self._runsFont['size']
        newFontSize = currentFontSize

        zoomPairs = [(32, 7),
                     (44, 6),
                     (53, 5),
                     (66, 4),
                     (73, 3),
                     (82, 2),
                     (90, 1),
                     (105, 0),
                     (120, -1),
                     (137, -2),
                     (999, -3),
                     ]

        for factor, sizeDecrement in zoomPairs:
            if self._zoomFactor <= factor:
                newFontSize = DEFAULT_FONT_SIZE - sizeDecrement
                break

        if currentFontSize != newFontSize:
            self._runsFont['size'] = newFontSize

    def zoomerP(self, event):
        self.__zoom(event, 1.111111)

    def zoomerM(self, event):
        self.__zoom(event, 0.9)

    def moveTo(self, x, y):

        if x > 1 or y > 1:
            x0,y0,x1,y1 = self.bbox("all")

            # x dim
            x = x / (x0+x1)
            start, end = self.xview()
            visisble_length = end - start
            x = x - (visisble_length/2)

            # Same with y
            y = y / (y0+y1)
            start, end = self.yview()
            visible_length = end - start
            y = y - (visible_length / 2)

        self.xview("moveto", x)
        self.yview("moveto", y)

    def drawGraph(self, graph, layout=None, drawNode=None, nodeList=None):
        """ Draw a graph in the canvas.
        nodes in the graph should have x and y.
        If layout is not None, it will be used to 
        reorganize the node positions.
        Provide drawNode if you want to customize how
        to create the boxes for each graph node.
        """

        # Reset the zoom and font
        scale = self._zoomFactor / DEFAULT_ZOOM
        self._zoomFactor = DEFAULT_ZOOM
        self._runsFont['size'] = DEFAULT_FONT_SIZE
        self.nodeList = nodeList

        if drawNode is None:
            self.drawNode = self._drawNode
        else:
            self.drawNode = drawNode

        self._drawNodes(graph.getRoot(), {})

        if layout is not None:
            layout.draw(graph)
            # Update node positions
            self._updatePositions(graph.getRoot(), {})

        self.updateScrollRegion()
        self.__zoom(None, scale)

    def reorganizeGraph(self, graph, layout=None):

        layout.draw(graph)
        # Update node positions
        self._updatePositions(graph.getRoot(), {}, createEdges=False)
        self.updateScrollRegion()

    def _drawNode(self, canvas, node):
        """ Default implementation to draw nodes as textboxes. """
        return TextBox(self, node.getLabel(), 0, 0,
                       bgColor="#99DAE8", textColor='black')

    def _drawNodes(self, node, visitedDict={}):
        nodeName = node.getName()

        if nodeName not in visitedDict:
            visitedDict[nodeName] = True
            item = self.drawNode(self, node)
            node.width, node.height = item.getDimensions()
            node.item = item
            item.node = node
            self.addItem(item)

            if getattr(node, 'expanded', True):
                for child in node.getChilds():
                    if self.nodeList is None:
                        self._drawNodes(child, visitedDict)
                    elif self.nodeList.getNode(child.run.getObjId()).isVisible():
                        self._drawNodes(child, visitedDict)
                    else:
                        self._setupParentProperties(node, visitedDict)
            else:
                self._setupParentProperties(node, visitedDict)

    def _connectParents(self, item):
        """
        Establishes a connection between the visible parents of node's children
        with node
        """
        visibleParents = self._visibleParents(item, [])
        for visibleParent in visibleParents:
            if visibleParent != item:
                dest = self.items[item.item.id]
                source = self.items[visibleParent.item.id]
                visibleParentNode = self.nodeList.getNode(visibleParent.run.getObjId())
                itemNode = self.nodeList.getNode(item.run.getObjId())

                if visibleParent not in item.getParents() and visibleParentNode.isExpanded():
                    self.createEdge(source, dest)
                if not itemNode.isExpanded():
                    self.createEdge(source, dest)

    def _visibleParents(self, node, parentlist):
        """
        Return a list with the visible parents of the node's children
        """
        for child in node.getChilds():
            parents = child.getParents()
            for parent in parents:
                parentNode = self.nodeList.getNode(parent.run.getObjId())
                if parentNode.isVisible() and parent != node and parent not in parentlist:
                        parentlist.append(parent)
        return parentlist

    def _setupParentProperties(self, node, visitedDict):
        """ This methods is used for collapsed nodes, in which 
        the properties (width, height, x and y) is propagated
        to the hidden childs.
        """
        for child in node.getChilds():
            if child.getName() not in visitedDict:
                child.width = node.width
                child.height = node.height
                child.x = node.x
                child.y = node.y
                self._setupParentProperties(child, visitedDict)

    def _updatePositions(self, node, visitedDict=None, createEdges=True):
        """ Update position of nodes and create the edges. """
        nodeName = node.getName()

        if nodeName not in visitedDict:
            visitedDict[nodeName] = True
            item = node.item
            item.moveTo(node.x, node.y)

            if getattr(node, 'expanded', True):
                for child in node.getChilds():
                    if self.nodeList is None:
                        self.createEdge(item, child.item)
                        self._updatePositions(child, visitedDict, createEdges)
                    elif self.nodeList.getNode(child.run.getObjId()).isVisible():
                        if createEdges:
                            self.createEdge(item, child.item)
                        self._updatePositions(child, visitedDict, createEdges)
            else:
                if createEdges:
                    self._connectParents(node)
                self._updatePositions(node, visitedDict, createEdges)


def findClosestPoints(list1, list2):
    candidates = []
    for c1 in list1:
        for c2 in list2:
            candidates.append([c1, c2, math.hypot(c2[0] - c1[0], c2[1] - c1[1])])
    closestTuple = min(candidates, key=operator.itemgetter(2))
    return closestTuple[0], closestTuple[1]


def findClosestConnectors(item1, item2):
    return findUpDownClosestConnectors(item1, item2)


def findUpDownClosestConnectors(item1, item2):
    srcConnectors = item1.getUpDownConnectorsCoordinates()
    dstConnectors = item2.getUpDownConnectorsCoordinates()
    if srcConnectors and dstConnectors:
        c1Coords, c2Coords = findClosestPoints(srcConnectors, dstConnectors)
        return c1Coords, c2Coords
    return None


def findStrictClosestConnectors(item1, item2):
    srcConnectors = item1.getConnectorsCoordinates()
    dstConnectors = item2.getConnectorsCoordinates()
    c1Coords, c2Coords = findClosestPoints(srcConnectors, dstConnectors)
    return c1Coords, c2Coords


def getConnectors(itemSource, itemDest):
    srcConnector = itemSource.getOutputConnectorCoordinates()
    dstConnector = itemDest.getInputConnectorCoordinates()

    return srcConnector, dstConnector


class Item(object):
    socketSeparation = 12
    verticalFlow = True

    TOP = 0
    RIGHT = 1
    BOTTOM = 2
    LEFT = 3

    def __init__(self, canvas, x, y):
        self.activeConnector = None
        self.canvas = canvas
        self.x = x
        self.y = y
        self.sockets = {}
        self.listeners = []
        self.selectionListeners = []
        self._selected = False

    def getCenter(self, x1, y1, x2, y2):
        xc = (x2 + x1) / 2.0
        yc = (y2 + y1) / 2.0
        return xc, yc

    def getConnectorsCoordinates(self):
        x1, y1, x2, y2 = self.getCorners()
        xc, yc = self.getCenter(x1, y1, x2, y2)
        return [(xc, y1), (x2, yc), (xc, y2), (x1, yc)]

    def getTopConnectorCoordinates(self):

        return self._getConnectorCoordinates(self.TOP)

    def getBottomConnectorCoordinates(self):

        return self._getConnectorCoordinates(self.BOTTOM)

    def getLeftConnectorCoordinates(self):

        return self._getConnectorCoordinates(self.LEFT)

    def getRightConnectorCoordinates(self):

        return self._getConnectorCoordinates(self.RIGHT)

    def _getConnectorCoordinates(self, side):
        fourConnectors = self.getConnectorsCoordinates()
        return fourConnectors[side]

    def getInputConnectorCoordinates(self):

        if self.verticalFlow:
            return self.getTopConnectorCoordinates()
        else:
            return self.getLeftConnectorCoordinates()

    def getOutputConnectorCoordinates(self):

        if self.verticalFlow:
            return self.getBottomConnectorCoordinates()
        else:
            return self.getRightConnectorCoordinates()

    def getUpDownConnectorsCoordinates(self):
        corners = self.getCorners()
        if corners:
            x1, y1, x2, y2 = self.getCorners()
            xc, yc = self.getCenter(x1, y1, x2, y2)
            return [(xc, y1), (xc, y2)]
        return None

    def getCorners(self):
        return self.canvas.bbox(self.id)

    def countSockets(self, verticalLocation):
        return len(list(self.getSocketsAt(verticalLocation)))

    def addSocket(self, name, socketClass, verticalLocation,
                  fillColor=DEFAULT_CONNECTOR_FILL,
                  outline=DEFAULT_CONNECTOR_OUTLINE,
                  position=None):
        count = self.countSockets(verticalLocation) + 1
        if position is None:
            position = count
        self.relocateSockets(verticalLocation, count)
        x, y = self.getSocketCoordsAt(verticalLocation, count, count)
        self.sockets[name] = {"object": socketClass(self.canvas, x, y, name,
                                                    fillColor=fillColor,
                                                    outline=outline),
                              "verticalLocation": verticalLocation,
                              "position": position}
        self.paintSocket(self.getSocket(name))

    def getSocket(self, name):
        return self.sockets[name]["object"]

    def getSocketsAt(self, verticalLocation):
        return filter(lambda s: s["verticalLocation"] == verticalLocation, self.sockets.values())

    def getSocketCoords(self, name):
        socket = self.sockets[name]
        return self.getSocketCoordsAt(socket["verticalLocation"], socket["position"],
                                      self.countSockets(socket["verticalLocation"]))

    def getSocketCoordsAt(self, verticalLocation, position=1, socketsCount=1):
        x1, y1, x2, y2 = self.getCorners()
        xc = (x2 + x1) / 2.0
        yc = (y1 + y2) / 2.0

        socketsGroupSize = (socketsCount - 1) * self.socketSeparation
        socketsGroupStart = xc - (socketsGroupSize / 2)
        x = socketsGroupStart + (position - 1) * self.socketSeparation
        if verticalLocation == "top":
            y = y1
        elif verticalLocation == 'bottom':
            y = y2
        elif verticalLocation == 'left':
            y = yc
            x = x1
        else:
            y = yc
            x = x2
        return x, y

    def relocateSockets(self, verticalLocation, count):
        sockets = self.getSocketsAt(verticalLocation)
        for socket in sockets:
            o = socket["object"]
            x, y = self.getSocketCoordsAt(verticalLocation, socket["position"], count)
            o.moveTo(x, y)

    def paintSocket(self, socket):
        # x,y=self.getSocketCoords(socket["name"])
        socket.paintSocket()

    def paintSockets(self):
        for name in self.sockets.keys():
            self.paintSocket(self.getSocket(name))

    def getDimensions(self):
        x, y, x2, y2 = self.canvas.bbox(self.id)
        return x2 - x, y2 - y

    def move(self, dx, dy):
        if hasattr(self, "id"):
            self.canvas.move(self.id, dx, dy)
        self.x += dx
        self.y += dy
        for name in self.sockets.keys():
            socket = self.sockets[name]
            socket["object"].move(dx, dy)
        for listenerFunc in self.listeners:
            listenerFunc(dx, dy)

    def moveTo(self, x, y):
        """Move TextBox to a new position (x,y)"""
        self.move(x - self.x, y - self.y)

    def addPositionListener(self, listenerFunc):
        self.listeners.append(listenerFunc)

    def addSelectionListener(self, listenerFunc):
        self.selectionListeners.append(listenerFunc)

    def _notifySelectionListeners(self, value):

        for listenerFunc in self.selectionListeners:
            listenerFunc(value)

    def setSelected(self, value):
        self._selected = value
        bw = 0
        bc = 'black'
        if value:
            bw = 2
            self.lift()
            # bc = 'Firebrick'
        else:
            self.lower()

        self.canvas.itemconfig(self.id, width=bw, outline=bc)
        self._notifySelectionListeners(value)

    def getSelected(self):
        return self._selected

    def lift(self):
        self.canvas.lift(self.id)

    def lower(self):
        self.canvas.lower(self.id)


class TextItem(Item):
    """This class will serve to paint and store rectangle boxes with some text.
       x and y are the coordinates of the center of this item"""

    def __init__(self, canvas, text, x, y, bgColor, textColor='black'):
        super(TextItem, self).__init__(canvas, x, y)
        self.bgColor = bgColor
        self.textColor = textColor
        self.text = text
        self.margin = 8
        self.paint()

    def _paintBounds(self, x, y, w, h, fillColor):
        """ Subclasses should implement this method 
        to paint the bounds to the text.
        Normally the bound are: rectangle or circles.
        Params:
            x, y: top left corner of the bounding box
            w, h: width and height of the box
            fillColor: color to fill the background
        Returns:
            should return the id of the created shape
        """
        pass

    def paint(self):
        """Paint the object in a specific position."""

        self.id_text = self.canvas.create_text(self.x, self.y, text=self.text,
                                               justify=tk.CENTER, fill=self.textColor,
                                               font=self.canvas.getRunsFont())

        xr, yr, w, h = self.canvas.bbox(self.id_text)
        m = self.margin

        self.id = self._paintBounds(xr - m, yr - m, w + m, h + m, fillColor=self.bgColor)
        self.canvas.lift(self.id_text)

    def move(self, dx, dy):
        super(TextItem, self).move(dx, dy)
        self.canvas.move(self.id_text, dx, dy)

    def lift(self):
        super(TextItem, self).lift()
        self.canvas.lift(self.id_text)

    def lower(self):
        self.canvas.lower(self.id_text)
        super(TextItem, self).lower()



class TextBox(TextItem):
    def __init__(self, canvas, text, x, y, bgColor, textColor='black'):
        super(TextBox, self).__init__(canvas, text, x, y, bgColor, textColor)

    def _paintBounds(self, x, y, w, h, fillColor):
        return self.canvas.create_rectangle(x, y, w, h, fill=fillColor, outline=fillColor)


class RoundedTextBox(TextItem):
    def __init__(self, canvas, text, x, y, bgColor, textColor='black'):
        super(RoundedTextBox, self).__init__(canvas, text, x, y, bgColor, textColor)

    def _paintBounds(self, upperLeftX, upperLeftY, bottomRightX, bottomRightY, fillColor):
        d = 5
        # When smooth=1, you define a straight segment by including its ends twice
        return self.canvas.create_polygon(upperLeftX + d + 1, upperLeftY,  # 1
                                          upperLeftX + d, upperLeftY,  # 1
                                          bottomRightX - d, upperLeftY,  # 2
                                          bottomRightX - d, upperLeftY,  # 2
                                          # bottomRightX-d+1,upperLeftY, #2b
                                          bottomRightX, upperLeftY + d - 1,  # 3b
                                          bottomRightX, upperLeftY + d,  # 3
                                          bottomRightX, upperLeftY + d,  # 3
                                          bottomRightX, bottomRightY - d,  # 4
                                          bottomRightX, bottomRightY - d,  # 4
                                          bottomRightX - d, bottomRightY,  # 5
                                          bottomRightX - d, bottomRightY,  # 5
                                          upperLeftX + d, bottomRightY,  # 6
                                          upperLeftX + d, bottomRightY,  # 6
                                          # upperLeftX+d-1,bottomRightY, #6b
                                          upperLeftX, bottomRightY - d + 1,  # 7b
                                          upperLeftX, bottomRightY - d,  # 7
                                          upperLeftX, bottomRightY - d,  # 7
                                          upperLeftX, upperLeftY + d,  # 8
                                          upperLeftX, upperLeftY + d,  # 8
                                          # upperLeftX, upperLeftY+d-1, #8b
                                          upperLeftX + d - 1, upperLeftY,  # 1b
                                          fill=fillColor, outline='black', smooth=1)

    def getDimensions(self):
        return self.canvas.bbox(self.id)


class TextCircle(TextItem):
    def __init__(self, canvas, text, x, y, bgColor, textColor='black'):
        super(TextCircle, self).__init__(canvas, text, x, y, bgColor, textColor)

    def _paintBounds(self, x, y, w, h, fillColor):
        return self.canvas.create_oval(x, y, w, h, fill=fillColor)


# This are not used and depends on xmippLib
# class ImageBox(Item):
#     def __init__(self, canvas, imgPath, x=0, y=0, text=None):
#         Item.__init__(self, canvas, x, y)
#         # Create the image
#         from pyworkflow.gui import getImage, getImageFromPath
#
#         if imgPath is None:
#             self.image = getImage('no-image.gif')
#         else:
#             self.image = getImageFromPath(imgPath)
#
#         if text is not None:
#             self.label = tk.Label(canvas, image=self.image, text=text,
#                                   compound=tk.TOP, bg='gray')
#             self.id = self.canvas.create_window(x, y, window=self.label)
#             self.label.bind('<Button-1>', self._onClick)
#         else:
#             self.id = self.canvas.create_image(x, y, image=self.image)
#
#
#     def setSelected(self, value): #Ignore selection highlight
#         pass
#
#     def _onClick(self, e=None):
#         pass


class Connector(Item):
    """ Default connector has no graphical representation (hence, it'ss invisible). Subclasses offer different looks"""

    def __init__(self, canvas, x, y, name):
        super(Connector, self).__init__(canvas, x, y)
        self.name = name

    def paintSocket(self):
        """Should be implemented by the subclasses"""
        pass

    def paintPlug(self, canvas, x, y):
        """Should be implemented by the subclasses"""
        pass

    def move(self, dx, dy):
        super(Connector, self).move(dx, dy)
        if hasattr(self, "socketId"):
            self.canvas.move(self.socketId, dx, dy)
        if hasattr(self, "plugId"):
            self.canvas.move(self.plugId, dx, dy)


class ColoredConnector(Connector):
    def __init__(self, canvas, x, y, name, fillColor=DEFAULT_CONNECTOR_FILL,
                 outline=DEFAULT_CONNECTOR_OUTLINE):
        super(ColoredConnector, self).__init__(canvas, x, y, name)
        self.fillColor = fillColor
        self.outline = outline


class RoundConnector(ColoredConnector):
    radius = 3

    def paintSocket(self):
        self.socketId = self.canvas.create_oval(self.x - self.radius,
                                                self.y - self.radius,
                                                self.x + self.radius,
                                                self.y + self.radius,
                                                outline=self.outline)

    def paintPlug(self):
        self.plugId = self.canvas.create_oval(self.x - self.radius,
                                              self.y - self.radius,
                                              self.x + self.radius,
                                              self.y + self.radius,
                                              fill=self.fillColor, width=0)


class SquareConnector(ColoredConnector):
    halfside = 3

    def paintSocket(self):
        self.socketId = self.canvas.create_rectangle(self.x - self.halfside,
                                                     self.y - self.halfside,
                                                     self.x + self.halfside,
                                                     self.y + self.halfside,
                                                     outline=self.outline)

    def paintPlug(self):
        self.plugId = self.canvas.create_rectangle(self.x - self.halfside,
                                                   self.y - self.halfside,
                                                   self.x + self.halfside,
                                                   self.y + self.halfside,
                                                   fill=self.fillColor, width=0)


# !!!! other figures: half circle, diamond...
class Oval:
    """Oval or circle"""

    def __init__(self, canvas, x, y, radio, color='green', anchor=None):
        self.anchor = anchor
        self.X, self.Y = x, y
        self.radio = radio
        self.color = color
        self.canvas = canvas
        anchor.addPositionListener(self.updateSrc)
        anchor.addSelectionListener(self.selectionListener)
        self.id = None
        self.paint()

    def paint(self):

        if self.id:
            self.canvas.delete(self.id)

        self.id = self.canvas.create_oval(self.X, self.Y, self.X + self.radio,
                                          self.Y + self.radio, fill=self.color,
                                          outline='black')
        # self.canvas.tag_raise(self.id)

    def updateSrc(self, dx, dy):
        self.X += dx
        self.Y += dy
        self.paint()

    def selectionListener(self, value):
        if value:
            self.canvas.lift(self.id)


class Rectangle:
    def __init__(self, canvas, x, y, width, height=None, color='green', anchor=None):

        self.anchor = anchor
        self.X, self.Y = x, y
        self.width = width
        self.height = height or width
        self.color = color
        self.canvas = canvas
        anchor.addPositionListener(self.updateSrc)
        anchor.addSelectionListener(self.selectionListener)
        self.id = None
        self.paint()

    def paint(self):

        if self.id:
            self.canvas.delete(self.id)

        self.id = self.canvas.create_rectangle(self.X, self.Y, self.X + self.width,
                                               self.Y + self.height, fill=self.color,
                                               outline=self.color)
        # self.canvas.tag_raise(self.id)

    def updateSrc(self, dx, dy):
        self.X += dx
        self.Y += dy
        self.paint()

    def selectionListener(self, value):
        if value:
            self.canvas.lift(self.id)


class Edge:
    """Edge between two objects"""

    def __init__(self, canvas, source, dest):
        self.source = source
        self.dest = dest
        self.srcX, self.srcY = source.x, source.y
        self.dstX, self.dstY = dest.x, dest.y
        self.canvas = canvas
        source.addPositionListener(self.updateSrc)
        dest.addPositionListener(self.updateDst)

        source.addSelectionListener(self.updateColor)
        dest.addSelectionListener(self.updateColor)
        self.id = None
        self.paint()

    def paint(self):
        # coords = findClosestConnectors(self.source,self.dest)
        coords = getConnectors(self.source, self.dest)

        if coords:
            c1Coords, c2Coords = coords

            if self.id:
                self.canvas.delete(self.id)

            lineColor = '#ccc'
            lineWidth = 1
            if not self.canvas.multipleItemsSelected():
                if self.dest.getSelected():
                    lineColor = '#000'
                    lineWidth = 2
                elif self.source.getSelected():
                    lineColor = '#b22222'
                    lineWidth = 2
            else:
                if self.dest.getSelected() and self.source.getSelected():
                    lineColor = '#000'
                    lineWidth = 2

            self.id = self.canvas.create_line(c1Coords[0], c1Coords[1],
                                              c2Coords[0], c2Coords[1],
                                              width=lineWidth, fill=lineColor)
            self.canvas.tag_lower(self.id)

    def updateSrc(self, dx, dy):
        self.srcX += dx
        self.srcY += dy
        self.paint()

    def updateDst(self, dx, dy):
        self.dstX += dx
        self.dstY += dy
        self.paint()

    def updateColor(self, value):
        self.paint()


# !!!! Interaction: allow to reconnect cables dynamically

# !!!! Antialiasing for the line - it seems TkInter does not support antialiasing...
# Although Tk 8.5 supports anti-aliasing if the Cairo library is installed:
# @see http://wiki.tcl.tk/10630

class Cable:
    def __init__(self, canvas, src, srcConnector, dst, dstConnector):
        self.id = None
        self.canvas = canvas
        self.srcPlug = src.getSocket(srcConnector)
        self.dstPlug = dst.getSocket(dstConnector)
        self.srcX = self.srcPlug.x
        self.srcY = self.srcPlug.y
        self.dstX, self.dstY = dst.getSocketCoords(dstConnector)
        src.addPositionListener(self.srcMoved)
        dst.addPositionListener(self.dstMoved)
        self.paint()

    def srcMoved(self, dx, dy):
        self.srcX = self.srcX + dx
        self.srcY = self.srcY + dy
        self.updateCoords()

    def updateCoords(self):
        self.canvas.coords(self.id, self.srcX, self.srcY, self.dstX, self.dstY)

    def dstMoved(self, dx, dy):
        self.dstX = self.dstX + dx
        self.dstY = self.dstY + dy
        self.updateCoords()

    def paint(self):
        self.id = self.canvas.create_line(self.srcX, self.srcY, self.dstX,
                                          self.dstY, width=2)
        self.canvas.tag_lower(self.id)
        self.paintPlugs()

    def paintPlugs(self):
        self.srcPlug.paintPlug()
        self.dstPlug.paintPlug()


if __name__ == '__main__':
    root = tk.Tk()
    canvas = Canvas(root, width=800, height=600)
    canvas.frame.grid(row=0, column=0, sticky='nsew')
    root.grid_columnconfigure(0, weight=1)
    root.grid_rowconfigure(0, weight=1)


    def canvasExample1():
        tb1 = canvas.createTextCircle("Project", 100, 100, "blue")
        tb2 = canvas.createTextbox("This is an intentionally quite big, big box,"
                                   "\nas you may appreciate looking carefully"
                                   "\nat it,\nas many times\nas you might need", 300, 200)
        tb2.addSocket("output1", RoundConnector, "bottom", fillColor="green")
        tb2.addSocket("output2", SquareConnector, "bottom", fillColor="yellow")
        tb2.addSocket("output3", SquareConnector, "bottom", fillColor="blue")
        tb3 = canvas.createRoundedTextbox("another one\n", 100, 200, "red")
        tb4 = canvas.createRoundedTextbox("tb4", 300, 300, "yellow")
        tb4.addSocket("input1", SquareConnector, "top", outline="red")
        tb5 = canvas.createTextCircle("tb5", 400, 300, "grey")
        tb5.addSocket("input1", SquareConnector, "top")
        e1 = canvas.createEdge(tb1, tb2)
        e2 = canvas.createEdge(tb1, tb3)
        c1 = canvas.createCable(tb2, "output2", tb4, "input1")
        c2 = canvas.createCable(tb2, "output3", tb5, "input1")
        tb3.moveTo(100, 300)


    canvasExample1()

    root.mainloop()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
###
"""
List all existing protocols within Scipion
"""

import sys

from pwem.protocols import (ProtImport, ProtMicrographs, ProtParticles, Prot2D,
                            Prot3D)
from pyworkflow import Config
from pyworkflow.viewer import Viewer
from pyworkflow.protocol.protocol import Protocol


def getFirstLine(doc):
    """ Get the first non empty line from doc. """
    if doc:
        for lines in doc.split('\n'):
            l = lines.strip()
            if l:
                return l
    return ''


def hasDoubleInheritance(classRef):
    # loop while class has single parent
    numParents = len(classRef.__bases__)
    while numParents == 1 and classRef is not Protocol:
        classRef = classRef.__bases__[0]
        numParents = len(classRef.__bases__)
    if numParents > 1:
        return True
    else:
        return False


if __name__ == '__main__':
    count = 0
    withDoc = '--with-doc' in sys.argv
    asciidoc = '--asciidoc' in sys.argv
    extended = '--extended' in sys.argv

    emProtocolsDict = Config.getDomain().getProtocols()
    emCategories = [('Imports', ProtImport, []),
                    ('Micrographs', ProtMicrographs, []),
                    ('Particles', ProtParticles, []),
                    ('2D', Prot2D, []),
                    ('3D', Prot3D, [])]
    protDict = {}

    # Group protocols by package name
    for k, v in emProtocolsDict.items():
        packageName = v.getClassPackageName()

        if packageName not in protDict:
            protDict[packageName] = []

        if not issubclass(v, Viewer) and not v.isBase():
            if extended:
                protTuple = (k, v, hasDoubleInheritance(v),
                             v().allowMpi, v().numberOfMpi,
                             v().allowThreads, v().numberOfThreads,
                             v().stepsExecutionMode)
            else:
                protTuple = (k, v)
            protDict[packageName].append(protTuple)
            for c in emCategories:
                if issubclass(v, c[1]):
                    c[2].append(protTuple)


    def iterGroups(protDict):
        groups = list(protDict.keys())
        groups.sort(key=lambda x: 1000-len(protDict[x]))

        for g in groups:
            yield g, protDict[g]

    def printProtocols(prots):
        protList = [(p[0], p[1], p[1].getClassLabel()) for p in prots]
        protList.sort(key=lambda x: x[2])

        for k, v, l in protList:
            doc = getFirstLine(v.__doc__) if withDoc else ''
            print("* link:%s[%s]: %s" % (k, l, doc))


    if asciidoc:
        print(":toc:\n:toc-placement!:\n\ntoc::[]\n")

        print("\n== By Categories\n")
        for c in emCategories:
            print("\n=== %s\n" % c[0])
            printProtocols(c[2])

        print("\n== By Packages\n")
        for group, prots in iterGroups(protDict):
            print("\n=== ", group, "(%d protocols)\n" % len(prots))
            printProtocols(prots)

    elif withDoc:
        for group, prots in iterGroups(protDict):
            print("Package: ", group, "(%d protocols)" % len(prots))
            for p in prots:
                print("   %s ( %s )" % (p[0], p[1].getClassLabel()))
                print("      doc: ", p[1].__doc__)
            print("-" * 100)

    else:
        if extended:
            formatStr = "{:<15}\t{:<35}\t{:<35}" + "\t{:<20}" * 6
            print(formatStr.format("PACKAGE", "PROTOCOL",
                                   "LABEL", "DOUBLE_INHERITANCE",
                                   "ALLOWS_MPI", "NUM_MPI",
                                   "ALLOWS_THREADS", "NUM_THREADS",
                                   "STEPS_EXEC_MODE"))
            for group, prots in iterGroups(protDict):
                for p in prots:
                    print(formatStr.format(group, p[0],
                                           p[1].getClassLabel(), *p[2:]))
        else:
            formatStr = "{:<15}\t{:<35}\t{:<35}"
            print(formatStr.format("PACKAGE", "PROTOCOL", "LABEL"))
            for group, prots in iterGroups(protDict):
                for k, v in prots:
                    print(formatStr.format(group, k, v.getClassLabel()))



# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es) [1]
# *              J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os

import pyworkflow as pw
import pyworkflow.object as pwobj
import pyworkflow.tests as pwtests
import pyworkflow.mapper as pwmapper
import pyworkflow.protocol as pwprot
from pyworkflow.project import Project
from pyworkflowtests.protocols import ProtOutputTest
from pyworkflowtests import Domain, MockObject


# Protocol to output of basic scipion objects
class TestProtocolOutputs(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls, writeLocalConfig=True)
        pwtests.setupTestOutput(cls)

        # Set the application domain
        pw.Config.setDomain("pyworkflowtests")

    def test_basicObjectOutput(self):
        """Test the list with several Complex"""
        fn = self.getOutputPath("protocol.sqlite")

        # Discover objects and protocols
        mapperDict = Domain.getMapperDict()

        mapper = pwmapper.SqliteMapper(fn, mapperDict)
        # Associate the project
        proj = Project(Domain, path=self.getOutputPath(''))

        prot = ProtOutputTest(mapper=mapper, n=2, project=proj,
                              workingDir=self.getOutputPath(''))

        # Add and old style o, not in the outputs dictionary
        prot.output1 = MockObject()

        self.assertFalse(prot._useOutputList.get(),
                         "useOutputList wrongly initialized")

        outputs = [o for o in prot.iterOutputAttributes()]
        self.assertTrue(1, len(outputs))

        prot._stepsExecutor = pwprot.StepExecutor(hostConfig=None)

        #Create the logs folder
        prot.makeWorkingDir()

        prot.run()

        self.assertEqual(prot._steps[0].getStatus(),
                         pwprot.STATUS_FINISHED)

        # Check there is an output
        self.assertOutput(prot)

        outputs = [o for o in prot.iterOutputAttributes()]

        # We are intentionally ignoring a protocol with o (EMObject)
        # That has been continued, We do not find a real case now.
        self.assertEqual(1, len(outputs),
                         msg="Integer o not registered properly.")

        outputs = [o for o in prot.iterOutputAttributes(pwobj.Integer)]

        # Test passing a filter
        self.assertEqual(1, len(outputs),
                         msg="Integer not matched when filtering outputs.")

        # Test with non existing class
        class NotRealClass:
            pass

        outputs = [o for o in prot.iterOutputAttributes(NotRealClass)]

        # Test passing a class
        self.assertEqual(0, len(outputs),
                         msg="Filter by class in iterOutputAttributes does "
                             "not work.")

        self.assertTrue(prot._useOutputList.get(),
                        "useOutputList not activated")

    def test_basicObjectInProject(self):
        prot = self.newProtocol(ProtOutputTest,
                                objLabel='to generate basic input')
        print("working dir: %s" % prot.getWorkingDir())
        # Define a negative output for later tests
        prot._defineOutputs(negative=pwobj.Integer(-20))
        self.launchProtocol(prot)

        # Default value is 10 so output is 20
        self.assertOutput(prot)

        # Second protocol to test linking
        prot2 = self.newProtocol(ProtOutputTest,
                                 objLabel='to read basic input')

        # Set the pointer for the integer
        prot2.iBoxSize.setPointer(pwobj.Pointer(prot, extended="oBoxSize"))
        self.launchProtocol(prot2)
        self.assertOutput(prot2, value=40)

        # Test validation: only positive numbers are allowed
        prot3 = self.newProtocol(ProtOutputTest,
                                 objLabel='invalid input',
                                 iBoxSize=-10)
        # We expect this to fail
        with self.assertRaises(Exception):
            self.launchProtocol(prot3)
        # Test validation: pointer value is validated
        prot4 = self.newProtocol(ProtOutputTest,
                                 objLabel='invalid pointer input')
        # Now use negative pointer output
        prot4.iBoxSize.setPointer(pwobj.Pointer(prot, extended="negative"))

        # We expect this to fail
        with self.assertRaises(Exception):
            self.launchProtocol(prot4)

    def assertOutput(self, prot, value=20):
        # Check there is an output
        self.assertTrue(hasattr(prot, 'oBoxSize'),
                        msg="Protocol output boxSize (OInteger) not registered"
                            " as attribute.")

        self.assertEqual(value, prot.oBoxSize.get(),
                         "oBoxSize value is wrong: %s , expected %s" %
                         (prot.oBoxSize, value))



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     Antonio Poza (Apr 30, 2013)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Module to handle default logging configuration and custom one. Default logging configuration
is defined here but optionally, it can be configured with an external json file
containing a standard python logging configuration content as documented here:
https://docs.python.org/3/howto/logging-cookbook.html#an-example-dictionary-based-configuration
To specify a custom logging config file use SCIPION_LOG_CONFIG variable to a json logging configuration file
If you just want to change the logger devel use SCIPION_LOG_LEVEL variable (defaults to INFO)
See https://docs.python.org/3/howto/logging.html#logging-levels for available levels. Use the literal! not de value.
"""
import os
import sys
import logging
import logging.config
import json
from logging import FileHandler

from pyworkflow.constants import PROJECT_SETTINGS, PROJECT_DBNAME
from pyworkflow.utils import Config

SCIPION_PROT_ID = "SCIPION_PROT_ID"
SCIPION_PROJ_ID = "SCIPION_PROJ_ID"


# Constant for extra logging data
class STATUS:
    START = "START"
    STOP = "STOP"
    INTERVAL = "INTERVAL"
    EVENT = "EVENT"


class LoggingConfigurator:
    """ Class to configure logging scenarios:

    1.- GUI logging

    2.- Protocol run logging"""

    customLoggingActive = False  # Holds if a custom logging configuration has taken place.

    @classmethod
    def setupLogging(cls):
        if not cls.loadCustomLoggingConfig():
            cls.setupDefaultLogging()

    @classmethod
    def loadCustomLoggingConfig(cls):
        """ Loads the custom logging configuration file"""
        from pyworkflow import Config

        if Config.SCIPION_LOG_CONFIG:
            if os.path.exists(Config.SCIPION_LOG_CONFIG):
                with open(Config.SCIPION_LOG_CONFIG, 'r') as stream:
                    config = json.load(stream)

                logging.config.dictConfig(config)
                cls.customLoggingActive = True
                return True
            else:
                print("SCIPION_LOG_CONFIG variable points to a non existing file: %s." % Config.SCIPION_LOG_CONFIG)
        return False

    @staticmethod
    def setupDefaultLogging():
        from pyworkflow import Config
        # Log configuration
        config = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'standard': {
                    'format': '%(asctime)s %(name)s %(levelname)s:  %(message)s'
                    # TODO: use formattime to show the time less verbose
                }
            },
            'handlers': {
                'fileHandler': {
                    'level': Config.SCIPION_LOG_LEVEL,
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'standard',
                    'filename': Config.SCIPION_LOG,
                    'maxBytes': 10000,
                    'backupCount': 10
                },
                'consoleHandler': {
                    'level': Config.SCIPION_LOG_LEVEL,
                    'class': 'logging.StreamHandler',
                    'formatter': 'standard',
                },
            },
            'loggers': {
                '': {
                    'handlers': ['consoleHandler', 'fileHandler'],
                    'level': Config.SCIPION_LOG_LEVEL,
                    'propagate': False,
                    'qualname': 'pyworkflow',
                },
            }
        }

        # Create the log folder
        os.makedirs(Config.SCIPION_LOGS, exist_ok=True)

        logging.config.dictConfig(config)

    @classmethod
    def setUpGUILogging(cls):
        """Sets up the logging library for the GUI processes: By default all goes to SCIPION_LOG file and console."""
        cls.setupLogging()

    @classmethod
    def setUpProtocolRunLogging(cls, stdoutLogFile, stderrLogFile):
        """ Sets up the logging library for the protocols run processes, loads the custom configuration plus
        2 FileHandlers for stdout and stderr"""

        stdoutHandler = FileHandler(stdoutLogFile)
        stderrHandler = FileHandler(stderrLogFile)
        stderrHandler.setLevel(logging.ERROR)

        # Get the roo logger
        rootLogger = logging.getLogger()

        # If there wasn't any custom logging
        if not cls.customLoggingActive:
            # Remove the default handler that goes to the terminal
            rootLogger.handlers.clear()

        # Add the 2 handlers, remove the
        rootLogger.addHandler(stderrHandler)
        rootLogger.addHandler(stdoutHandler)
        rootLogger.setLevel(Config.SCIPION_LOG_LEVEL)

        if Config.debugOn():
            # create formatter and add it to the handlers
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s <-- %(name)s')
            stdoutHandler.setFormatter(formatter)
            stderrHandler.setFormatter(formatter)

        # Capture std out and std err and send it to the file handlers
        rootLogger.info("Logging configured. STDOUT --> %s , STDERR --> %s" % (stdoutLogFile, stderrLogFile))
        sys.stderr = stderrHandler.stream
        sys.stdout = stdoutHandler.stream

        return rootLogger


def restoreStdoutAndErr():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


# ******** Extra code to send some log lines to an external performance analysis tool ***********
def setDefaultLoggingContext(protId, projId):
    os.environ[SCIPION_PROT_ID] = str(protId)
    os.environ[SCIPION_PROJ_ID] = projId


def getFinalProtId(protId):
    return protId if protId is not None else int(os.environ.get(SCIPION_PROT_ID, "-1"))


def getFinalProjId(projId):
    return projId if projId is not None else os.environ.get(SCIPION_PROJ_ID, "unknown")


def getExtraLogInfo(measurement, status, project_name=None, prot_id=None, prot_name=None, step_id=None, duration=None,
                    dbfilename=None):
    try:
        # Add TS!! optionally
        if dbfilename:
            splitDb = dbfilename.split("/")
            dbName = splitDb[-1]
            runName = ""
            # project.sqlite and settings.sqlite may not have elements
            if dbName not in [PROJECT_SETTINGS, PROJECT_DBNAME]:
                runName = splitDb[1]
            dbfilename = os.path.join(runName, dbName)

        return {"measurement": measurement,
                "status": status,
                "project_name": getFinalProjId(project_name),
                "prot_id": getFinalProtId(prot_id),
                "prot_name": prot_name,
                "step_id": step_id,
                "duration": duration,
                "dbfilename": dbfilename
                }

    except Exception as e:
        print("getExtraLogInfo failed: %s" % e)



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This script is intended to be invoked using ssh from a master.
So, it will just do the same job of launch._launchLocal and print
the jobId to be tracked from the machine that was invoked.
"""
import os
import sys
import subprocess

import pyworkflow as pw
from pyworkflow.protocol import getProtocolFromDb
from pyworkflow.protocol.launch import launch, stop


def usage(msg=''):
    print("Usage: pw_protocol_remote.py [run|stop] project protDbPath protID\n"
          "%s" % msg)
    sys.exit(1)
    
    
if __name__ == '__main__':
    n = len(sys.argv)
    if n < 5:
        usage("Received only %d arguments" % n)
        
    mode = sys.argv[1]
    
    if mode not in ['run', 'stop']:
        usage("Mode should be 'run' or 'stop'. Received: '%s'" % mode)

    projectPath = os.path.join(pw.Config.SCIPION_USER_DATA, 'projects',
                               sys.argv[2])
    print("projectPath: ", projectPath)
    protDbPath = sys.argv[3]
    protId = int(sys.argv[4])

    protocol = getProtocolFromDb(projectPath, protDbPath, protId, chdir=False)
    # We need to change the hostname to localhost, since it will 
    # be considered 'local' from now on to either run or stop
    protocol.setHostName('localhost')
    
    if mode == 'run':        
        FNULL = open(os.devnull, 'w')
        jobId = launch(protocol, stdin=None, stdout=FNULL,
                       stderr=subprocess.STDOUT)
        print("Scipion remote jobid: %d" % jobId)
    elif mode == 'stop':
        stop(protocol)
    else:
        usage()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import contextlib
import sys
import os
import re
from datetime import datetime
import traceback
from enum import Enum

import bibtexparser
import numpy as np
import math

from pyworkflow import Config


def prettyDate(time=False):
    """
    Get a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc
    """
    now = datetime.now()
    if type(time) is int:
        diff = now - datetime.fromtimestamp(time)
    elif type(time) is float:
        diff = now - datetime.fromtimestamp(int(time))
    elif isinstance(time, datetime):
        diff = now - time
    elif not time:
        # Avoid now - now (sonar cloud bug)
        copy = now
        diff = now - copy
    second_diff = diff.seconds
    day_diff = diff.days

    if day_diff < 0:
        return ''

    if day_diff == 0:
        if second_diff < 10:
            return "just now"
        if second_diff < 60:
            return str(second_diff) + " seconds ago"
        if second_diff < 120:
            return "a minute ago"
        if second_diff < 3600:
            return str(int(second_diff / 60)) + " minutes ago"
        if second_diff < 7200:
            return "an hour ago"
        if second_diff < 86400:
            return str(int(second_diff / 3600)) + " hours ago"
    if day_diff == 1:
        return "Yesterday"
    if day_diff < 7:
        return str(day_diff) + " days ago"
    if day_diff < 31:
        return str(int(day_diff / 7)) + " weeks ago"
    if day_diff < 365:
        return str(int(day_diff / 30)) + " months ago"
    return str(int(day_diff / 365)) + " years ago"


def dateStr(dt=None, time=True, secs=False, dateFormat=None):
    """ Get a normal string representation of datetime. 
    If dt is None, use NOW.
    """
    if dt is None:
        dt = datetime.now()
    elif isinstance(dt, float) or isinstance(dt, int):
        dt = datetime.fromtimestamp(dt)

    if dateFormat is None:
        dateFormat = '%d-%m-%Y'
        if time:
            dateFormat += ' %H:%M'
            if secs:
                dateFormat += ':%S'

    return dt.strftime(dateFormat)


prettyTime = dateStr


def prettyTimestamp(dt=None, format='%Y-%m-%d_%H%M%S'):
    if dt is None:
        dt = datetime.now()

    return dt.strftime(format)


def prettySize(size):
    """ Human friendly file size. """
    unit_list = list(zip(['bytes', 'kB', 'MB', 'GB', 'TB', 'PB'],
                         [0, 0, 1, 2, 2, 2]))
    if size > 1:
        exponent = min(int(math.log(size, 1024)), len(unit_list) - 1)
        quotient = float(size) / 1024 ** exponent
        unit, num_decimals = unit_list[exponent]
        format_string = '{:.%sf} {}' % num_decimals
        return format_string.format(quotient, unit)
    if size == 0:
        return '0 bytes'
    if size == 1:
        return '1 byte'


def prettyDelta(timedelta):
    """ Remove the milliseconds of the timedelta. """
    return str(timedelta).split('.')[0]

class UtcConverter:
    """ Class to make date conversions to utc"""
    utc_delta = datetime.utcnow() - datetime.now()

    def __call__(cls, t):
        return t + cls.utc_delta
# Use to_utc like a function: to_utc(date)
to_utc = UtcConverter()

def prettyLog(msg):
    print(cyan(prettyTime(datetime.now(), secs=True)), msg)


class Timer(object):
    """ Simple Timer base in datetime.now and timedelta. """

    def __init__(self, message=""):
        self._message = message

    def tic(self):
        self._dt = datetime.now()

    def getElapsedTime(self):
        return datetime.now() - self._dt

    def toc(self, message='Elapsed:'):
        print(message, self.getElapsedTime())

    def getToc(self):
        return prettyDelta(self.getElapsedTime())

    def __enter__(self):
        self.tic()

    def __exit__(self, type, value, traceback):
        self.toc(self._message)


def timeit(func):
    """ Decorator function to have a simple measurement
    of the execution time of a given function.
    Just use:
    @timeit
    def func(...)
        ...
    to use it.
    """

    def timedFunc(*args, **kwargs):
        t = Timer()
        t.tic()
        result = func(*args, **kwargs)
        t.toc("Function '%s' took" % func)

        return result

    return timedFunc


def trace(nlevels, separator=' --> ', stream=sys.stdout):
    # Example:
    #   @trace(3)
    #   def doRefresh(...
    # gives as output whenever doRefresh is called lines like:
    #   text.py:486 _addFileTab --> text.py:330 __init__ --> doRefresh

    def realTrace(f):
        """ Decorator function to print stack call in a human-readable way.
        """

        def tracedFunc(*args, **kwargs):
            stack = traceback.extract_stack()[-nlevels - 1:-1]
            fmt = lambda x: '%s:%d %s' % (os.path.basename(x[0]), x[1], x[2])
            stList = list(map(fmt, stack))
            stream.write(separator.join(stList + [f.__name__]) + '\n')
            return f(*args, **kwargs)

        return tracedFunc

    return realTrace


def prettyDict(d):
    print("{")
    for k, v in d.items():
        print("    %s: %s" % (k, v))
    print("}")


def prettyXml(elem, level=0):
    """ Add indentation for XML elements for more human readable text. """
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for _elem in elem:
            prettyXml(_elem, level + 1)
        if not _elem.tail or not _elem.tail.strip():
            _elem.tail = i


def getUniqueItems(originalList):
    """ Method to remove repeated items from one list 
    originalList -- Original list with repeated items, or not.
    returns -- New list with the content of original list without repeated items
    """
    auxDict = {}
    resultList = [auxDict.setdefault(x, x) for x in originalList if x not in auxDict]
    return resultList


def executeRemoteX(command, hostName, userName, password):
    """ Execute a remote command with X11 forwarding.
    Params:
        command: Command to execute.
        hostName: Remote host name.
        userName: User name.
        password: Password.
    Returns: 
        Tuple with standard output and error output.
    """
    scriptPath = os.path.abspath(os.path.join(os.path.dirname(__file__), "sshAskPass.sh"))
    pswCommand = "echo '" + password + "' | " + scriptPath + " ssh -X " + userName + "@" + hostName + " " + command
    import subprocess
    p = subprocess.Popen(pswCommand, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    return stdout, stderr


def executeRemote(command, hostName, userName, password):
    """ Execute a remote command.
    Params:
        command: Command to execute.
        hostName: Remote host name.
        userName: User name.
        password: Password.
    Returns: 
        Tuple with standard input, standard output and error output.
    """
    import paramiko
    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostName, 22, userName, password)
    stdin, stdout, stderr = ssh.exec_command(command)
    ssh.close()
    return stdin, stdout, stderr


def executeLongRemote(command, hostName, userName, password):
    """ Execute a remote command.
    Params:
        command: Command to execute.
        hostName: Remote host name.
        userName: User name.
        password: Password.
    Returns: 
        Tuple with standard input, standard output and error output.
    """
    import paramiko
    import select
    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostName, 22, userName, password)
    transport = ssh.get_transport()
    channel = transport.open_session()
    channel.exec_command(command)
    while True:
        if channel.exit_status_ready():
            break
        rl, wl, xl = select.select([channel], [], [], 0.0)
        if len(rl) > 0:
            print(channel.recv(1024))


def getLocalUserName():
    """ Recover local machine user name.
    returns: Local machine user name.
    """
    import getpass
    return getpass.getuser()


def getLocalHostName():
    return getHostName()


def getHostName():
    """ Return the name of the local machine. """
    import socket
    return socket.gethostname()


def getHostFullName():
    """ Return the fully-qualified name of the local machine. """
    import socket
    return socket.getfqdn()


# ******************************File utils *******************************

def isInFile(text, filePath):
    """ Checks if given text is in the given file.
    params:
        text: Text to check.
        filePath : File path to check.
    returns: True if the given text is in the given file, 
             False if it is not in the file.
    """
    return any(text in line for line in open(filePath))


def getLineInFile(text, fileName):
    """ Find the line where the given text is located in the given file.
    params:
       text: Text to check.
       filePath : File path to check.
    returns: File number where the text was located.
    """
    with open(fileName) as f:
        for i, line in enumerate(f):
            if text in line:
                return i + 1
    return None

def hasAnyFileChanged(files, time):
    """ Returns true if any of the files in files list has been changed after 'time'"""
    for file in files:
        if hasFileChangedSince(file, time):
            return True

    return False

def hasFileChangedSince(file, time):
    """ Returns true if the file has changed after 'time'"""
    modTime = datetime.datetime.fromtimestamp(getmtime(file))
    return time < modTime


# ------------- Colored message strings -----------------------------

class StrColors(Enum):
    gray = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    magenta = 35
    cyan = 36


def getColorStr(text, color, bold=False):
    """ Add ANSI color codes to the string if there is a terminal sys.stdout.
    Params:
     text: text to be colored
     color: red or green
     bold: bold the text
    """
    if envVarOn('SCIPION_SAFE_COLORS') and not sys.stdout.isatty():
        return text

    attr = [str(color.value)]

    if bold:
        attr.append('1')
    return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), text)


def grayStr(text):
    return getColorStr(text, color=StrColors.gray)


def redStr(text):
    return getColorStr(text, color=StrColors.red)


def greenStr(text):
    return getColorStr(text, color=StrColors.green)


def yellowStr(text):
    return getColorStr(text, color=StrColors.yellow)


def blueStr(text):
    return getColorStr(text, color=StrColors.blue)


def magentaStr(text):
    return getColorStr(text, color=StrColors.magenta)


def cyanStr(text):
    return getColorStr(text, color=StrColors.cyan)


def ansi(n, bold=False):
    """Return function that escapes text with ANSI color n."""
    return lambda txt: '\x1b[%d%sm%s\x1b[0m' % (n, ';1' if bold else '', txt)


black, red, green, yellow, blue, magenta, cyan, white = map(ansi, range(30, 38))
blackB, redB, greenB, yellowB, blueB, magentaB, cyanB, whiteB = [
    ansi(i, bold=True) for i in range(30, 38)]

# -------------- Hyper text highlighting ----------------------------
"""
We use a subset of TWiki hyper text conventions.
In particular:
    *some_text* will display some_text in bold
    _some_text_ will display some_text in italic
    Links:
        http://www.link-page.com  -> hyperlink using the url as label
        [[http://www.link-page.com][Link page]] -> hyperlink using "Link page" as label
"""
# Types of recognized styles
HYPER_BOLD = 'bold'
HYPER_ITALIC = 'italic'
HYPER_LINK1 = 'link1'
HYPER_SCIPION_OPEN = 'sci-open'
HYPER_LINK2 = 'link2'
HYPER_ALL = 'all'

# Associated regular expressions
PATTERN_BOLD = "(^|[\s])[*](?P<bold>[^\s*][^*]*[^\s*]|[^\s*])[*]"
# PATTERN_BOLD = r"[\s]+[*]([^\s][^*]+[^\s])[*][\s]+"
PATTERN_ITALIC = "(^|[\s])[_](?P<italic>[^\s_][^_]*[^\s_]|[^\s_])[_]"
# PATTERN_ITALIC = r"[\s]+[_]([^\s][^_]+[^\s])[_][\s]+"
PATTERN_LINK1 = '(?P<link1>http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+#]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)'
PATTERN_LINK2 = "[\[]{2}(?P<link2>[^\s][^\]]+[^\s])[\]][\[](?P<link2_label>[^\s][^\]]+[^\s])[\]]{2}"
# __PATTERN_LINK2 should be first since it could contain __PATTERN_LINK1
PATTERN_ALL = '|'.join([PATTERN_BOLD, PATTERN_ITALIC, PATTERN_LINK2, PATTERN_LINK1])

# Compiled regex
# Not need now, each pattern compiled separately
# HYPER_REGEX = {
#               HYPER_BOLD: re.compile(PATTERN_BOLD),
#               HYPER_ITALIC: re.compile(PATTERN_ITALIC),
#               HYPER_LINK1: re.compile(PATTERN_LINK1),
#               HYPER_LINK2: re.compile(PATTERN_LINK1),
#               }
HYPER_ALL_RE = re.compile(PATTERN_ALL)


def parseHyperText(text, matchCallback):
    """ Parse the text recognizing Hyper definitions below.
    Params:
        matchCallback: a callback function to processing each matching,
                       it should accept the type of match (HYPER_BOLD, ITALIC or LINK)
    Return:
        The input text with the replacements made by matchCallback
    """

    def _match(match):
        """ Call the proper matchCallback with some extra info. """
        m = match.group().strip()
        if m.startswith('*'):
            tag = HYPER_BOLD
        elif m.startswith('_'):
            tag = HYPER_ITALIC
        elif m.startswith('http'):
            tag = HYPER_LINK1
        elif m.startswith('[['):
            tag = HYPER_LINK2
        else:
            raise Exception("Bad prefix for HyperText match")
        return matchCallback(match, tag)

    return HYPER_ALL_RE.sub(_match, text)


#    for hyperMode, hyperRegex in HYPER_REGEX.iteritems():
#        text = hyperRegex.sub(lambda match: matchCallback(match, hyperMode), text)
#
#    return text

class LazyDict(object):
    """ Dictionary to be initialized in the moment it is accessed for the first time.
    Initialization is done by a callback passed at instantiation"""
    def __init__(self, callback=dict):
        """ :param callback: method to initialize the dictionary. SHould return a dictionary"""
        self.data = None
        self.callback = callback

    def evaluate_callback(self):
        self.data = self.callback()

    def __getitem__(self, name):
        if self.data is None:
            self.evaluate_callback()
        return self.data.__getitem__(name)

    def __setitem__(self, name, value):
        if self.data is None:
            self.evaluate_callback()
        return self.data.__setitem__(name, value)

    def __getattr__(self, name):
        if self.data is None:
            self.evaluate_callback()
        return getattr(self.data, name)

    def __iter__(self):
        if self.data is None:
            self.evaluate_callback()
        return self.data.__iter__()


def parseBibTex(bibtexStr):
    """ Parse a bibtex file and return a dictionary. """

    return bibtexparser.loads(bibtexStr).entries_dict



def isPower2(num):
    """ Return True if 'num' is a power of 2. """
    return num != 0 and ((num & (num - 1)) == 0)


# ---------------------------------------------------------------------------
# Parsing of arguments
# ---------------------------------------------------------------------------

def getListFromRangeString(rangeStr):
    """ Create a list of integers from a string with range definitions.
    Examples:
    "1,5-8,10" -> [1,5,6,7,8,10]
    "2,6,9-11" -> [2,6,9,10,11]
    "2 5, 6-8" -> [2,5,6,7,8]
    """
    # Split elements by command or space
    elements = re.split(',|\s', rangeStr)
    values = []
    for e in elements:
        if '-' in e:
            limits = e.split('-')
            values += range(int(limits[0]), int(limits[1]) + 1)
        else:
            # If values are separated by comma also splitted 
            values += map(int, e.split())
    return values


def getRangeStringFromList(list):
    left = None
    right = None
    ranges = []

    def addRange():
        if left == right:  # Single element
            ranges.append("%d" % right)
        else:
            ranges.append("%(left)d-%(right)d" % locals())

    for item in list:
        if right is None:
            left = right = item
        else:
            if item == right + 1:
                right += 1
            else:
                addRange()
                left = right = item
    addRange()
    return ','.join(ranges)


def getListFromValues(valuesStr, length=None):
    """ Convert a string representing list items into a list.
    The items should be separated by spaces and a multiplier 'x' can be used.
    If length is not None, then the last element will be repeated
    until the desired length is reached.
    Examples:
    '1 1 2x2 4 4' -> ['1', '1', '2', '2', '4', '4']
    '2x3, 3x4, 1' -> ['3', '3', '4', '4', '4', '1']
    """
    result = []

    for chunk in valuesStr.split():
        values = chunk.split('x')
        n = len(values)
        if n == 1:  # 'x' is not present in the chunk, single value
            result += values
        elif n == 2:  # multiple the values by the number after 'x'
            result += [values[1]] * int(values[0])
        else:
            raise Exception("More than one 'x' is not allowed in list string value.")

    # If length is passed, we fill the list with 
    # the last element until length is reached
    if length is not None and length > len(result):
        item = result[-1]
        result += [item] * (length - len(result))

    return result


def getFloatListFromValues(valuesStr, length=None):
    """ Convert a string to a list of floats"""
    return [float(v) for v in getListFromValues(valuesStr, length)]


def getBoolListFromValues(valuesStr, length=None):
    """ Convert a string to a list of booleans"""
    from pyworkflow.object import Boolean
    return [Boolean(value=v).get() for v in getListFromValues(valuesStr, length)]


def getStringListFromValues(valuesStr, length=None):
    """ Convert a string to a list of booleans"""
    from pyworkflow.object import String
    return [String(value=v).get() for v in getListFromValues(valuesStr, length)]


class Environ(dict):
    """ Some utilities to handle environment settings. """
    REPLACE = 0
    BEGIN = 1
    END = 2

    def getFirst(self, keys, mandatory=False):
        """ Return the value of the first key present in the environment.
        If none is found, returns the 'defaultValue' parameter.
        """
        for k in keys:
            if k in self:
                return self.get(k)

        if mandatory:
            print("None of the variables: %s found in the Environment. "
                  "Please check scipion.conf files." % (str(keys)))

        return None

    def set(self, varName, varValue, position=REPLACE):
        """ Modify the value for some variable.
        Params:
            varName: for example LD_LIBRARY_PATH
            varValue: the value to add or replace.
            position: controls how the value will be changed.
                If REPLACE, it will overwrite the value of
                the var.
                BEGIN or END will preserve the current value
                and add (at begin or end) the new value.
        """
        if varName in self and position != self.REPLACE:
            if position == self.BEGIN:
                self[varName] = varValue + os.pathsep + self[varName]
            elif position == self.END:
                self[varName] = self[varName] + os.pathsep + varValue
        else:
            self[varName] = varValue

    def update(self, valuesDict, position=REPLACE):
        """ Use set for each key, value pair in valuesDict. """
        for k, v in valuesDict.items():
            self.set(k, v, position)

    def addLibrary(self, libraryPath, position=BEGIN):
        """ Adds a path to LD_LIBRARY_PATH at the requested position
        if the provided paths exist. """

        if libraryPath is None:
            return

        if existsVariablePaths(libraryPath):
            self.update({'LD_LIBRARY_PATH': libraryPath}, position=position)
        else:
            print("Some paths do not exist in: % s" % libraryPath)

    def setPrepend(self, prepend):
        """ Use this method to set a prepend string that will be added at
        the beginning of any command that will be run in this environment.
        This can be useful for example when 'modules' need to be loaded and
        a simple environment variables setup is not enough.
        """
        setattr(self, '__prepend', prepend)

    def getPrepend(self):
        """ Return if there is any prepend value. See setPrepend function. """
        return getattr(self, '__prepend', '')


def existsVariablePaths(variableValue):
    """ Check if the path (or paths) in variableValue exists.
    Multiple paths are allowed if separated by os."""
    return all(os.path.exists(p)
               for p in variableValue.split(os.pathsep) if p.split())


def environAdd(varName, newValue, valueFirst=False):
    """ Add a new value to some environ variable.
    If valueFirst is true, the new value will be at the beginning.
    """
    varList = [os.environ[varName]]
    i = 1
    if valueFirst:
        i = 0
    varList.insert(i, newValue)
    os.environ[varName] = os.pathsep.join(varList)


def envVarOn(varName, env=None):
    """ Is variable set to True in the environment? """
    v = env.get(varName) if env else os.environ.get(varName)
    return strToBoolean(v)

def strToBoolean(string):
    return string is not None and string.lower() in ['true', 'yes', 'on', '1']


def getMemoryAvailable():
    """ Return the total memory of the system in MB """
    from psutil import virtual_memory
    return virtual_memory().total // 1024 ** 2


def startDebugger(password='a'):
    if Config.debugOn():
        try:
            # FIXME: rpdb2 does not support python 3
            from rpdb2 import start_embedded_debugger
            print("Starting debugger...")
            start_embedded_debugger(password)
        except Exception:
            print("Error importing rpdb2 debugging module, consider installing winpdb.")


def getFreePort(basePort=0, host=''):
    import socket
    port = 0
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, basePort))
        ipaddr, port = s.getsockname()
        s.close()
    except Exception as e:
        print(e)
        return 0
    return port


def readProperties(propsFile):
    myprops = {}
    with open(propsFile, 'r') as f:
        for line in f:
            line = line.rstrip()  # removes trailing whitespace and '\n' chars

            if "=" not in line:
                continue  # skips blanks and comments w/o =
            if line.startswith("#"):
                continue  # skips comments which contain =

            k, v = line.split("=", 1)
            myprops[k] = v
    return myprops


# ---------------------Color utils --------------------------
def hex_to_rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))


def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb


def lighter(color, percent):
    """assumes color is rgb between (0, 0, 0) and (255, 255, 255)"""
    color = np.array(color)
    white = np.array([255, 255, 255])
    vector = white - color
    return tuple(np.around(color + vector * percent))


def formatExceptionInfo(level=6):
    error_type, error_value, trbk = sys.exc_info()
    tb_list = traceback.format_tb(trbk, level)
    s = "Error: %s \nDescription: %s \nTraceback:" % (error_type.__name__,
                                                      error_value)
    for i in tb_list:
        s += "\n" + i
    return s


def printTraceBack():
    traceback.print_stack()


def getEnvVariable(variableName, default=None, exceptionMsg=None):
    """ Returns the value of an environment variable or raise an exception message.
    Useful when adding variable to the config file and report accurate messages"""
    value = os.getenv(variableName)

    if exceptionMsg is None:
        exceptionMsg = ("Environment variable %s not found. "
                        "Please check scipion configuration. "
                        "Try running : scipion config." % variableName)

    if value is None:
        if default is None:
            raise Exception(exceptionMsg)
        else:
            return default
    else:
        return value


@contextlib.contextmanager
def weakImport(package):
    """
    This method can be use to tolerate imports that may fail, e.g imports

    from .protocol_ctffind import CistemProtCTFFind
    with weakImport('tomo'):
        from .protocol_ts_ctffind import CistemProtTsCtffind

    in this case CistemProtTsCtffind should fail if tomo package is missing,
    but exception is captured and all the imports above should be available

    :param package: name of the package that is expected to fail
    """
    try:
        yield
    except ImportError as e:
        if "'%s'" % package not in str(e):
            raise e



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Pablo Conesa [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This modules hosts code provider and window to search for a protocol"""
import tkinter as tk

from pyworkflow import Config
import  pyworkflow.gui as pwgui
import pyworkflow.object as pwobj
from pyworkflow.gui.dialog import SearchBaseWindow

from pyworkflow.gui.project.utils import isAFinalProtocol
from pyworkflow.gui.project.viewprotocols_extra import ProtocolTreeConfig


class ProtocolTreeProvider(pwgui.tree.ObjectTreeProvider):
    """Create the tree elements for a Protocol run"""

    def __init__(self, protocol):
        self.protocol = protocol
        # This list is create to group the protocol parameters
        # in the tree display
        self.status = pwobj.List(objName='_status')
        self.params = pwobj.List(objName='_params')
        self.statusList = ['status', 'initTime', 'endTime', 'error',
                           'interactive', 'mode']

        objList = [] if protocol is None else [protocol]
        pwgui.tree.ObjectTreeProvider.__init__(self, objList)



class SearchProtocolWindow(SearchBaseWindow):

    columnConfig = {
        '#0': ('Status', {'width': 50, 'minwidth': 50, 'stretch': tk.NO}, 5),
        # Heading, tree column kwargs, casting for sorting
        'protocol': ('Protocol', {'width': 300, 'stretch': tk.FALSE}, 10),
        'streaming': ('Streamified', {'width': 100, 'stretch': tk.FALSE}, 5),
        'installed': ('Installation', {'width': 110, 'stretch': tk.FALSE}, 5),
        'help': ('Help', {'minwidth': 300, 'stretch': tk.YES}, 5),
        'score': ('Score', {'width': 50, 'stretch': tk.FALSE}, 5, int)
    }

    def __init__(self, parentWindow, **kwargs):

        super().__init__(parentWindow,
                         title="Add a protocol",
                         **kwargs)

    def _createResultsTree(self, frame, show, columns):
        return self.master.getViewWidget()._createProtocolsTree(frame, show=show, columns=columns)

    def _onSearchClick(self, e=None):

        self._resultsTree.clear()

        protList = self.scoreProtocols()

        # Sort by weight
        protList.sort(reverse=True, key=lambda x: x[7])

        self._addProtocolToTree(protList)

    def scoreProtocols(self):

        keyword = self._searchVar.get().lower().strip()
        emProtocolsDict = Config.getDomain().getProtocols()
        protList = []

        for key, prot in emProtocolsDict.items():
            if isAFinalProtocol(prot, key):
                label = prot.getClassLabel().lower()
                line = (key, label,
                        "installed" if prot.isInstalled() else "missing installation",
                        prot.getHelpText().strip().replace('\r', '').replace('\n', '').lower(),
                        "streamified" if prot.worksInStreaming() else "static",
                        "beta" if prot.isBeta() else "",
                        "new" if prot.isNew() else "")

                line = self._addSearchWeight(line, keyword)
                # something was found: weight > 0
                if line[7] != 0:
                    protList.append(line)

        return protList

    @staticmethod
    def _addSearchWeight(line2Search, searchtext):
        # Adds a weight value for the search
        weight = 0

        # prioritize findings in label
        if searchtext in line2Search[1]:
            weight += 10

        for value in line2Search[2:]:
            weight += 5 if searchtext in value else 0

        if " " in searchtext:
            for word in searchtext.split():
                if word in line2Search[1]:
                    weight += 3

                for value in line2Search[2:]:
                    weight += 1 if word in value else 0

        return line2Search + (weight,)

    def _addProtocolToTree(self, protList):
        """ Adds the items in protList to the tree

        :param protList: List of tuples with all the values/colunms used in search ans shown in the tree"""

        for key, label, installed, help, streamified, beta, new, weight in protList:
            tag = ProtocolTreeConfig.getProtocolTag(installed == 'installed',
                                                    beta == 'beta',
                                                    new == 'new')
            self._resultsTree.insert(
                '', 'end', key, text="", tags=tag,
                values=(label, streamified, installed, help, weight))





# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.tests as pwtests

from pyworkflow import Config
from pyworkflowtests.protocols import  ConcurrencyProtocol


class TestConcurrency(pwtests.BaseTest):
    
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestOutput(cls)

    
       # Set the application domain
        Config.setDomain("pyworkflowtests")
        pwtests.setupTestProject(cls)


    def test_simple_steps_concurrency(self):
        prot = self.newProtocol(ConcurrencyProtocol, numberOfThreads=3)

        self.launchProtocol(prot)
